from iman import *


def fun(fname , b):
    ffmpeg_command =  '%s -y -i "%s" -vf "scale=w=%s:h=%s" -compression_level %s "%s"' %(b[3] ,fname , b[0],b[1] ,b[5], PJ(b[4],PN(fname)+b[2]) )
    cmd(ffmpeg_command)
   
   
 
def convert (fname_pattern ,ext ='jpg',ofolder=None , w=-1 , h=-1,level=100,  worker=4,ffmpeg_path='c:\\ffmpeg.exe'): 
    
    if (not ext.startswith('.')):
       ext = '.' + ext    
       
    files = gf(fname_pattern) 
    

    
    if (len(files)==0):
        print('no File!!!')
        return
    
   
    if (ofolder==None):
        pp = input('It is better to set ofolder\n press y to continue conversion: ')
        if (pp.lower()=='y'):
          ofolder=""
        else:
           return        
    else:
       PM(ofolder)
     


    worker=min(len(files) , worker)     

    from multiprocessing import Pool
    from functools import partial  
    with Pool(processes=worker) as pool:
         pool.map(partial(fun, b=[w , h,ext , ffmpeg_path,ofolder,level]), files)     


def fun2(fname , b):
    ffmpeg_command =  '%s -y -i "%s" -vf "scale=iw/%s:ih/%s" "%s"' %(b[3] ,fname , b[0],b[1] , PJ(b[4],PN(fname)+b[2]) )
    cmd(ffmpeg_command)
   
   
 
def resize (fname_pattern ,ext ='jpg',ofolder=None , w=2 , h=2,  worker=4,ffmpeg_path='c:\\ffmpeg.exe'): 
    
    if (not ext.startswith('.')):
       ext = '.' + ext    
       
    files = gf(fname_pattern) 
    

    
    if (len(files)==0):
        print('no File!!!')
        return
    
   
    if (ofolder==None):
        pp = input('It is better to set ofolder\n press y to continue conversion: ')
        if (pp.lower()=='y'):
          ofolder=""
        else:
           return        
    else:
       PM(ofolder)
     


    worker=min(len(files) , worker)     

    from multiprocessing import Pool
    from functools import partial  
    with Pool(processes=worker) as pool:
         pool.map(partial(fun2, b=[w , h,ext , ffmpeg_path,ofolder]), files)     



# -*- coding: utf-8 -*-
"""
Created on Wed Jun 10 14:09:03 2020

@author: user
"""
#Extract MFCC with Librosa

import numpy as np
from scipy import signal
from .getmfcc_from_librosa import *


def framing(sig, win_size, win_shift=1, context=(0, 0), pad='zeros'):
    """
    :param sig: input signal, can be mono or multi dimensional
    :param win_size: size of the window in term of samples
    :param win_shift: shift of the sliding window in terme of samples
    :param context: tuple of left and right context
    :param pad: can be zeros or edge
    """
    dsize = sig.dtype.itemsize
    if sig.ndim == 1:
        sig = sig[:, np.newaxis]
    # Manage padding
    c = (context, ) + (sig.ndim - 1) * ((0, 0), )
    _win_size = win_size + sum(context)
    shape = (int((sig.shape[0] - win_size) / win_shift) + 1, 1, _win_size, sig.shape[1])
    strides = tuple(map(lambda x: x * dsize, [win_shift * sig.shape[1], 1, sig.shape[1], 1]))
    if pad == 'zeros':
        return np.lib.stride_tricks.as_strided(np.lib.pad(sig, c, 'constant', constant_values=(0,)),
                                                  shape=shape,
                                                  strides=strides).squeeze()
    elif pad == 'edge':
        return np.lib.stride_tricks.as_strided(np.lib.pad(sig, c, 'edge'),
                                                  shape=shape,
                                                  strides=strides).squeeze()
    
    
def preemphasis(wav, coeff=0.975):
    preem_wav = signal.lfilter([1, -coeff], [1], wav)
    return preem_wav

def wav2melspec(wav, sr, n_fft, win_length, hop_length, n_mels, time_first=True, **kwargs):
    # Linear spectrogram
    mag_spec, phase_spec = wav2spec(wav, n_fft, win_length, hop_length, time_first=False)

    # Mel-spectrogram
    mel_spec = linear_to_mel(mag_spec, sr, n_fft, n_mels, **kwargs)

    # Time-axis first
    if time_first:
        mel_spec = mel_spec.T  # (t, n_mels)

    return mel_spec

def wav2melspec_db(wav, sr, n_fft, win_length, hop_length, n_mels,time_first=True, **kwargs):
    # Mel-spectrogram
    mel_spec = wav2melspec(wav, sr, n_fft, win_length, hop_length, n_mels, time_first=False, **kwargs)

    # Decibel
    mel_db = amplitude_to_db(mel_spec)

    # Time-axis first
    if time_first:
        mel_db = mel_db.T  # (t, n_mels)

    return mel_db

def wav2spec(wav, n_fft, win_length, hop_length, time_first=True):
    stft = STFT(y=wav, n_fft=n_fft, hop_length=hop_length, win_length=win_length)
    mag = np.abs(stft)
    phase = np.angle(stft)

    if time_first:
        mag = mag.T
        phase = phase.T

    return mag, phase

def wav2mfcc(wav, sr, n_fft=256, win_length=200, hop_length=80, n_mels=40, n_mfccs=24, preemphasis_coeff=0.975, time_first=True,
             **kwargs):
    # Pre-emphasis
    wav_preem = preemphasis(wav, coeff=preemphasis_coeff)
    
    
    framed = framing(wav, win_length, win_shift=hop_length).copy()
    framed = preemphasis(framed, 0.97)
   
  
    log_energy = np.log((framed**2).sum(axis=1))
    

    # Decibel-scaled mel-spectrogram
    mel_db = wav2melspec_db(wav_preem, sr, n_fft, win_length, hop_length, n_mels, time_first=False, **kwargs)

    # MFCCs
    mfccs = np.dot(dct(n_mfccs, mel_db.shape[0]), mel_db)


    # Time-axis first
    if time_first:
        mfccs = mfccs.T  # (t, n_mfccs)

    return np.asarray(mfccs , dtype = np.float32),log_energy

def linear_to_mel(linear, sr, n_fft, n_mels, **kwargs):
    mel_basis = MEL(sr, n_fft, n_mels, **kwargs)  # (n_mels, 1+n_fft//2)
    mel = np.dot(mel_basis, linear)  # (n_mels, t) # mel spectrogram
    return mel
    
def local_MVN(X , num):
    feat=np.empty((0,np.shape(X)[1]))
    window = num*2
    le = len(X)//window
    for i in range(le+1):
       xx = X[(i*window) : ((i+1)*window)]
       if (len(xx)==0):
          break
       m = xx.mean(axis=0)
       s = xx.std(axis=0)
       xx_mvn = (xx-m)/(s+(2**-30))
       feat=np.insert(feat,i*window,xx_mvn , axis=0)
    return(feat)
 

def Get(wav,sample_rate,le=False):
     mfccs,log_energy = wav2mfcc (wav , sr=sample_rate)
     if (le):
         return mfccs,log_energy
     else:
         return mfccs
         
def Normal(mfccs , win_len=150):
   mfccs = local_MVN(mfccs,win_len)
   return(mfccs)



from multiprocessing import Pool
from functools import partial
d=[1,2,3,4,5,6,7,8,9]

def fun2(fname):
    print(fname)

def fun(dd,b):
   print(b[0])

if (__name__ == '__main__'):
  with Pool(processes=4) as pool:
     fea = pool.map(partial(fun, b=['Sons',8000]), d)
	 
  with Pool(processes=4) as pool:
       pool.map(fun , d)


# -*- coding: utf-8 -*-
"""
Created on Wed Apr 10 11:08:34 2019

@author: deepspeech
"""

import numpy as np
import soundfile as sf
import pyworld

wav ,fs = sf.read(r'rooh.wav' ,dtype='float64' )


f0, timeaxis = pyworld.harvest(wav, fs, frame_period = 10, f0_floor = 71.0, f0_ceil = 800.0)
sp = pyworld.cheaptrick(wav, f0, timeaxis, fs)
ap = pyworld.d4c(wav, f0, timeaxis, fs)

# Get Mel-cepstral coefficients (MCEPs)
#coded_sp = pyworld.code_spectral_envelope(sp, fs, 24)
#fftlen = pyworld.get_cheaptrick_fft_size(fs)
#decoded_sp = pyworld.decode_spectral_envelope(coded_sp, fs, fftlen)



wav2 = pyworld.synthesize(f0*2, sp*5, ap*5, fs,12)

wav2 = wav2.astype(np.float32)
sf.write(r'rooh2.wav', wav2, 16000, format='wav', subtype='PCM_16')




from iman import *

def help():
   print('\nneed to install winrar\n')
   
   
def rar(fname , out="" , rar_path=r"C:\Program Files\WinRAR\winrar.exe"):

    if (out==""):
        out = PJ(PD(fname) , PN(fname))
    else:
        out =  PJ(PD(out) , PN(out))   
        
    command = '"%s" a -r "%s.rar" "%s" ' %(rar_path , out , fname)
    cmd(command)
     
def unrar(fname , out="" , rar_path=r"C:\Program Files\WinRAR\winrar.exe"):
    
    if (PE(fname)!=".rar"):
        print("Just Extract rar files")
        return
    if (out!=""):
       out = out + "\\"
       command = '"%s" x -ad -c- -cfg- -inul -o+ -y "%s" "%s" ' %(rar_path , fname , out)
    else:
        command = '"%s" x -ad -c- -cfg- -inul -o+ -y "%s"' %(rar_path , fname )  
    cmd(command)
  
   
def zip(fname , out="" , rar_path=r"C:\Program Files\WinRAR\winrar.exe"):

    if (out==""):
        out = PJ(PD(fname) , PN(fname))
    else:
        out =  PJ(PD(out) , PN(out))   
        
    command = '"%s" a -afzip -r "%s.zip" "%s" ' %(rar_path , out , fname)
    cmd(command)
   
def unzip(fname , out="" , rar_path=r"C:\Program Files\WinRAR\winrar.exe"):
    
    if (PE(fname)!=".zip"):
        print("Just Extract zip files")
        return
    if (out!=""):
       out = out + "\\"
       command = '"%s" x -ibck "%s" "%s" ' %(rar_path , fname , out)
    else:
         command = '"%s" x -ibck "%s"' %(rar_path , fname)  
    cmd(command)
   
   
   
   


import numpy as np
import six
import scipy
import scipy.fftpack as fft
from numpy.lib.stride_tricks import as_strided

MAX_MEM_BLOCK = 2**8 * 2**10

class LibrosaError(Exception):
    '''The root librosa exception class'''
    pass


class ParameterError(LibrosaError):
    '''Exception class for mal-formed inputs'''
    pass
 



def tiny(x):

    # Make sure we have an array view
    x = np.asarray(x)

    # Only floating types generate a tiny
    if np.issubdtype(x.dtype, np.floating) or np.issubdtype(
        x.dtype, np.complexfloating
    ):
        dtype = x.dtype
    else:
        dtype = np.float32

    return np.finfo(dtype).tiny



 
def expand_to(x, *, ndim, axes):
   

    # Force axes into a tuple

    try:
        axes = tuple(axes)
    except TypeError:
        axes = tuple([axes])

    if len(axes) != x.ndim:
        raise ParameterError(
            "Shape mismatch between axes={} and input x.shape={}".format(axes, x.shape)
        )

    if ndim < x.ndim:
        raise ParameterError(
            "Cannot expand x.shape={} to fewer dimensions ndim={}".format(x.shape, ndim)
        )

    shape = [1] * ndim
    for i, axi in enumerate(axes):
        shape[axi] = x.shape[i]

    return x.reshape(shape)



def fix_length(data, *, size, axis=-1, **kwargs):
    
    kwargs.setdefault("mode", "constant")

    n = data.shape[axis]

    if n > size:
        slices = [slice(None)] * data.ndim
        slices[axis] = slice(0, size)
        return data[tuple(slices)]

    elif n < size:
        lengths = [(0, 0)] * data.ndim
        lengths[axis] = (0, size - n)
        return np.pad(data, lengths, **kwargs)

    return data


    
def dtype_c2r(d, *, default=np.float32):
    mapping = {
        np.dtype(np.complex64): np.float32,
        np.dtype(np.complex128): np.float64,
        np.dtype(complex): np.dtype(np.float32).type,   
    }

    # If we're given a real type already, return it
    dt = np.dtype(d)
    if dt.kind == "f":
        return dt

    # Otherwise, try to map the dtype.
    # If no match is found, return the default.
    return np.dtype(mapping.get(np.dtype(d), default))    
def __overlap_add(y, ytmp, hop_length):
    # numba-accelerated overlap add for inverse stft
    # y is the pre-allocated output buffer
    # ytmp is the windowed inverse-stft frames
    # hop_length is the hop-length of the STFT analysis

    n_fft = ytmp.shape[-2]
    for frame in range(ytmp.shape[-1]):
        sample = frame * hop_length
        y[..., sample : (sample + n_fft)] += ytmp[..., frame]
        
  
def normalize(S, *, norm=np.inf, axis=0, threshold=None, fill=None):
   

    # Avoid div-by-zero
    if threshold is None:
        threshold = tiny(S)

    elif threshold <= 0:
        raise ParameterError(
            "threshold={} must be strictly " "positive".format(threshold)
        )

    if fill not in [None, False, True]:
        raise ParameterError("fill={} must be None or boolean".format(fill))

    if not np.all(np.isfinite(S)):
        raise ParameterError("Input must be finite")

    # All norms only depend on magnitude, let's do that first
    mag = np.abs(S).astype(float)

    # For max/min norms, filling with 1 works
    fill_norm = 1

    if norm == np.inf:
        length = np.max(mag, axis=axis, keepdims=True)

    elif norm == -np.inf:
        length = np.min(mag, axis=axis, keepdims=True)

    elif norm == 0:
        if fill is True:
            raise ParameterError("Cannot normalize with norm=0 and fill=True")

        length = np.sum(mag > 0, axis=axis, keepdims=True, dtype=mag.dtype)

    elif np.issubdtype(type(norm), np.number) and norm > 0:
        length = np.sum(mag ** norm, axis=axis, keepdims=True) ** (1.0 / norm)

        if axis is None:
            fill_norm = mag.size ** (-1.0 / norm)
        else:
            fill_norm = mag.shape[axis] ** (-1.0 / norm)

    elif norm is None:
        return S

    else:
        raise ParameterError("Unsupported norm: {}".format(repr(norm)))

    # indices where norm is below the threshold
    small_idx = length < threshold

    Snorm = np.empty_like(S)
    if fill is None:
        # Leave small indices un-normalized
        length[small_idx] = 1.0
        Snorm[:] = S / length

    elif fill:
        # If we have a non-zero fill value, we locate those entries by
        # doing a nan-divide.
        # If S was finite, then length is finite (except for small positions)
        length[small_idx] = np.nan
        Snorm[:] = S / length
        Snorm[np.isnan(Snorm)] = fill_norm
    else:
        # Set small values to zero by doing an inf-divide.
        # This is safe (by IEEE-754) as long as S is finite.
        length[small_idx] = np.inf
        Snorm[:] = S / length

    return Snorm
    
def get_window(window, Nx, *, fftbins=True):
   
    if callable(window):
        return window(Nx)

    elif isinstance(window, (str, tuple)) or np.isscalar(window):
        # TODO: if we add custom window functions in librosa, call them here

        return scipy.signal.get_window(window, Nx, fftbins=fftbins)

    elif isinstance(window, (np.ndarray, list)):
        if len(window) == Nx:
            return np.asarray(window)

        raise ParameterError(
            "Window size mismatch: " "{:d} != {:d}".format(len(window), Nx)
        )
    else:
        raise ParameterError("Invalid window specification: {}".format(window))

    
def __window_ss_fill(x, win_sq, n_frames, hop_length):  # pragma: no cover
    """Helper function for window sum-square calculation."""

    n = len(x)
    n_fft = len(win_sq)
    for i in range(n_frames):
        sample = i * hop_length
        x[sample : min(n, sample + n_fft)] += win_sq[: max(0, min(n_fft, n - sample))]
    
def window_sumsquare(
    *,
    window,
    n_frames,
    hop_length=512,
    win_length=None,
    n_fft=2048,
    dtype=np.float32,
    norm=None,
):
   
    if win_length is None:
        win_length = n_fft

    n = n_fft + hop_length * (n_frames - 1)
    x = np.zeros(n, dtype=dtype)

    # Compute the squared window at the desired length
    win_sq = get_window(window, win_length)
    win_sq = normalize(win_sq, norm=norm) ** 2
    win_sq = pad_center(win_sq, size=n_fft)

    # Fill the envelope
    __window_ss_fill(x, win_sq, n_frames, hop_length)

    return x

    
def istft(
    stft_matrix,
    *,
    hop_length=None,
    win_length=None,
    n_fft=None,
    window="hann",
    center=True,
    dtype=None,
    length=None,
):
    
    if n_fft is None:
        n_fft = 2 * (stft_matrix.shape[-2] - 1)

    # By default, use the entire frame
    if win_length is None:
        win_length = n_fft

    # Set the default hop, if it's not already specified
    if hop_length is None:
        hop_length = int(win_length // 4)

    ifft_window = get_window(window, win_length, fftbins=True)

    # Pad out to match n_fft, and add broadcasting axes
    ifft_window = pad_center(ifft_window, size=n_fft)
    ifft_window = expand_to(ifft_window, ndim=stft_matrix.ndim, axes=-2)

    # For efficiency, trim STFT frames according to signal length if available
    if length:
        if center:
            padded_length = length + int(n_fft)
        else:
            padded_length = length
        n_frames = min(stft_matrix.shape[-1], int(np.ceil(padded_length / hop_length)))
    else:
        n_frames = stft_matrix.shape[-1]

    if dtype is None:
        dtype = dtype_c2r(stft_matrix.dtype)

    shape = list(stft_matrix.shape[:-2])
    expected_signal_len = n_fft + hop_length * (n_frames - 1)
    shape.append(expected_signal_len)
    y = np.zeros(shape, dtype=dtype)

    n_columns = MAX_MEM_BLOCK // (
        np.prod(stft_matrix.shape[:-1]) * stft_matrix.itemsize
    )
    n_columns = max(n_columns, 1)

    from numpy import fft

    frame = 0
    for bl_s in range(0, n_frames, n_columns):
        bl_t = min(bl_s + n_columns, n_frames)

        # invert the block and apply the window function
        ytmp = ifft_window * fft.irfft(stft_matrix[..., bl_s:bl_t], n=n_fft, axis=-2)

        # Overlap-add the istft block starting at the i'th frame
        __overlap_add(y[..., frame * hop_length :], ytmp, hop_length)

        frame += bl_t - bl_s

    # Normalize by sum of squared window
    ifft_window_sum = window_sumsquare(
        window=window,
        n_frames=n_frames,
        win_length=win_length,
        n_fft=n_fft,
        hop_length=hop_length,
        dtype=dtype,
    )

    approx_nonzero_indices = ifft_window_sum > tiny(ifft_window_sum)
    y[..., approx_nonzero_indices] /= ifft_window_sum[approx_nonzero_indices]

    if length is None:
        # If we don't need to control length, just do the usual center trimming
        # to eliminate padded data
        if center:
            y = y[..., int(n_fft // 2) : -int(n_fft // 2)]
    else:
        if center:
            # If we're centering, crop off the first n_fft//2 samples
            # and then trim/pad to the target length.
            # We don't trim the end here, so that if the signal is zero-padded
            # to a longer duration, the decay is smooth by windowing
            start = int(n_fft // 2)
        else:
            # If we're not centering, start at 0 and trim/pad as necessary
            start = 0

        y = fix_length(y[..., start:], size=length)

    return y


def pad_center(data, size, axis=-1, **kwargs):

    kwargs.setdefault('mode', 'constant')

    n = data.shape[axis]

    lpad = int((size - n) // 2)

    lengths = [(0, 0)] * data.ndim
    lengths[axis] = (lpad, int(size - n - lpad))

    if lpad < 0:
        raise ParameterError(('Target size ({:d}) must be '
                              'at least input size ({:d})').format(size, n))

    return np.pad(data, lengths, **kwargs)

def valid_audio(y, mono=True):
    if not isinstance(y, np.ndarray):
        raise ParameterError('data must be of type numpy.ndarray')

    if not np.issubdtype(y.dtype, np.floating):
        raise ParameterError('data must be floating-point')

    if mono and y.ndim != 1:
        raise ParameterError('Invalid shape for monophonic audio: '
                             'ndim={:d}, shape={}'.format(y.ndim, y.shape))

    elif y.ndim > 2 or y.ndim == 0:
        raise ParameterError('Audio must have shape (samples,) or (channels, samples). '
                             'Received shape={}'.format(y.shape))

    if not np.isfinite(y).all():
        raise ParameterError('Audio buffer is not finite everywhere')

    return True
	
def frame(y, frame_length=2048, hop_length=512):

    if not isinstance(y, np.ndarray):
        raise ParameterError('Input must be of type numpy.ndarray, '
                             'given type(y)={}'.format(type(y)))

    if y.ndim != 1:
        raise ParameterError('Input must be one-dimensional, '
                             'given y.ndim={}'.format(y.ndim))

    if len(y) < frame_length:
        raise ParameterError('Buffer is too short (n={:d})'
                             ' for frame_length={:d}'.format(len(y), frame_length))

    if hop_length < 1:
        raise ParameterError('Invalid hop_length: {:d}'.format(hop_length))

    if not y.flags['C_CONTIGUOUS']:
        raise ParameterError('Input buffer must be contiguous.')

    # Compute the number of frames that will fit. The end may get truncated.
    n_frames = 1 + int((len(y) - frame_length) / hop_length)

    # Vertical stride is one sample
    # Horizontal stride is `hop_length` samples
    y_frames = as_strided(y, shape=(frame_length, n_frames),
                          strides=(y.itemsize, hop_length * y.itemsize))
    return y_frames
	
def power_to_db(S, ref=1.0, amin=1e-10, top_db=80.0):
  

    S = np.asarray(S)

    if amin <= 0:
        raise ParameterError('amin must be strictly positive')

    if np.issubdtype(S.dtype, np.complexfloating):
        print('power_to_db was called on complex input so phase '
                      'information will be discarded. To suppress this warning, '
                      'call power_to_db(magphase(D, power=2)[0]) instead.')
        magnitude = np.abs(S)
    else:
        magnitude = S

    if six.callable(ref):
        # User supplied a function to calculate reference power
        ref_value = ref(magnitude)
    else:
        ref_value = np.abs(ref)

    log_spec = 10.0 * np.log10(np.maximum(amin, magnitude))
    log_spec -= 10.0 * np.log10(np.maximum(amin, ref_value))

    if top_db is not None:
        if top_db < 0:
            raise ParameterError('top_db must be non-negative')
        log_spec = np.maximum(log_spec, log_spec.max() - top_db)

    return log_spec	


def amplitude_to_db(S, ref=1.0, amin=1e-5, top_db=80.0):
    S = np.asarray(S)

    if np.issubdtype(S.dtype, np.complexfloating):
        print('amplitude_to_db was called on complex input so phase '
                      'information will be discarded. To suppress this warning, '
                      'call amplitude_to_db(magphase(D)[0]) instead.')

    magnitude = np.abs(S)

    if six.callable(ref):
        # User supplied a function to calculate reference power
        ref_value = ref(magnitude)
    else:
        ref_value = np.abs(ref)

    power = np.square(magnitude, out=magnitude)

    return power_to_db(power, ref=ref_value**2, amin=amin**2,
                       top_db=top_db)
					   
					   
def STFT(y, n_fft=2048, hop_length=None, win_length=None, window='hann',
         center=True, dtype=np.complex64, pad_mode='reflect'):
    # By default, use the entire frame
    if win_length is None:
        win_length = n_fft

    # Set the default hop, if it's not already specified
    if hop_length is None:
        hop_length = int(win_length // 4)

    fft_window = get_window(window, win_length, fftbins=True)

    # Pad the window out to n_fft size
    fft_window = pad_center(fft_window, n_fft)

    # Reshape so that the window can be broadcast
    fft_window = fft_window.reshape((-1, 1))

    # Check audio is valid
    valid_audio(y)

    # Pad the time series so that frames are centered
    if center:
        y = np.pad(y, int(n_fft // 2), mode=pad_mode)

    # Window the time series.
    y_frames = frame(y, frame_length=n_fft, hop_length=hop_length)

    # Pre-allocate the STFT matrix
    stft_matrix = np.empty((int(1 + n_fft // 2), y_frames.shape[1]),
                           dtype=dtype,
                           order='F')

    # how many columns can we fit within MAX_MEM_BLOCK?
    n_columns = int(MAX_MEM_BLOCK / (stft_matrix.shape[0] *
                                          stft_matrix.itemsize))

    for bl_s in range(0, stft_matrix.shape[1], n_columns):
        bl_t = min(bl_s + n_columns, stft_matrix.shape[1])

        # RFFT and Conjugate here to match phase from DPWE code
        stft_matrix[:, bl_s:bl_t] = fft.fft(fft_window *
                                            y_frames[:, bl_s:bl_t],
                                            axis=0)[:stft_matrix.shape[0]]

    return stft_matrix


def get_window(window, Nx, fftbins=True):
    if six.callable(window):
        return window(Nx)

    elif (isinstance(window, (six.string_types, tuple)) or
          np.isscalar(window)):
        # TODO: if we add custom window functions in librosa, call them here

        return scipy.signal.get_window(window, Nx, fftbins=fftbins)

    elif isinstance(window, (np.ndarray, list)):
        if len(window) == Nx:
            return np.asarray(window)

        raise ParameterError('Window size mismatch: '
                             '{:d} != {:d}'.format(len(window), Nx))
    else:
        raise ParameterError('Invalid window specification: {}'.format(window))

def hz_to_mel(frequencies, htk=False):

    frequencies = np.asanyarray(frequencies)

    if htk:
        return 2595.0 * np.log10(1.0 + frequencies / 700.0)

    # Fill in the linear part
    f_min = 0.0
    f_sp = 200.0 / 3

    mels = (frequencies - f_min) / f_sp

    # Fill in the log-scale part

    min_log_hz = 1000.0                         # beginning of log region (Hz)
    min_log_mel = (min_log_hz - f_min) / f_sp   # same (Mels)
    logstep = np.log(6.4) / 27.0                # step size for log region

    if frequencies.ndim:
        # If we have array data, vectorize
        log_t = (frequencies >= min_log_hz)
        mels[log_t] = min_log_mel + np.log(frequencies[log_t]/min_log_hz) / logstep
    elif frequencies >= min_log_hz:
        # If we have scalar data, heck directly
        mels = min_log_mel + np.log(frequencies / min_log_hz) / logstep

    return mels

def mel_to_hz(mels, htk=False):
   
    mels = np.asanyarray(mels)

    if htk:
        return 700.0 * (10.0**(mels / 2595.0) - 1.0)

    # Fill in the linear scale
    f_min = 0.0
    f_sp = 200.0 / 3
    freqs = f_min + f_sp * mels

    # And now the nonlinear scale
    min_log_hz = 1000.0                         # beginning of log region (Hz)
    min_log_mel = (min_log_hz - f_min) / f_sp   # same (Mels)
    logstep = np.log(6.4) / 27.0                # step size for log region

    if mels.ndim:
        # If we have vector data, vectorize
        log_t = (mels >= min_log_mel)
        freqs[log_t] = min_log_hz * np.exp(logstep * (mels[log_t] - min_log_mel))
    elif mels >= min_log_mel:
        # If we have scalar data, check directly
        freqs = min_log_hz * np.exp(logstep * (mels - min_log_mel))

    return freqs


def fft_frequencies(sr=22050, n_fft=2048):


    return np.linspace(0,
                       float(sr) / 2,
                       int(1 + n_fft//2),
                       endpoint=True)

def mel_frequencies(n_mels=128, fmin=0.0, fmax=11025.0, htk=False):

    # 'Center freqs' of mel bands - uniformly spaced between limits
    min_mel = hz_to_mel(fmin, htk=htk)
    max_mel = hz_to_mel(fmax, htk=htk)

    mels = np.linspace(min_mel, max_mel, n_mels)

    return mel_to_hz(mels, htk=htk)

def MEL(sr, n_fft, n_mels=128, fmin=0.0, fmax=None, htk=False,
        norm=1):

    if fmax is None:
        fmax = float(sr) / 2

    if norm is not None and norm != 1 and norm != np.inf:
        raise ParameterError('Unsupported norm: {}'.format(repr(norm)))

    # Initialize the weights
    n_mels = int(n_mels)
    weights = np.zeros((n_mels, int(1 + n_fft // 2)))

    # Center freqs of each FFT bin
    fftfreqs = fft_frequencies(sr=sr, n_fft=n_fft)

    # 'Center freqs' of mel bands - uniformly spaced between limits
    mel_f = mel_frequencies(n_mels + 2, fmin=fmin, fmax=fmax, htk=htk)

    fdiff = np.diff(mel_f)
    ramps = np.subtract.outer(mel_f, fftfreqs)

    for i in range(n_mels):
        # lower and upper slopes for all bins
        lower = -ramps[i] / fdiff[i]
        upper = ramps[i+2] / fdiff[i+1]

        # .. then intersect them with each other and zero
        weights[i] = np.maximum(0, np.minimum(lower, upper))

    if norm == 1:
        # Slaney-style mel is scaled to be approx constant energy per channel
        enorm = 2.0 / (mel_f[2:n_mels+2] - mel_f[:n_mels])
        weights *= enorm[:, np.newaxis]

    # Only check weights if f_mel[0] is positive
    if not np.all((mel_f[:-2] == 0) | (weights.max(axis=1) > 0)):
        # This means we have an empty channel somewhere
        print('Empty filters detected in mel frequency basis. '
                      'Some channels will produce empty responses. '
                      'Try increasing your sampling rate (and fmax) or '
                      'reducing n_mels.')

    return weights

def dct(n_filters, n_input):
    basis = np.empty((n_filters, n_input))
    basis[0, :] = 1.0 / np.sqrt(n_input)

    samples = np.arange(1, 2*n_input, 2) * np.pi / (2.0 * n_input)

    for i in range(1, n_filters):
        basis[i, :] = np.cos(i*samples) * np.sqrt(2.0/n_input)

    return basis


import json

default_param = {}
default_param['bins'] = -1
default_param['unstable_bins'] = -1 # training only
default_param['stable_bins'] = -1 # training only
default_param['sr'] = 44100
default_param['pre_filter_start'] = -1
default_param['pre_filter_stop'] = -1
default_param['band'] = {}

N_BINS = 'n_bins'

def int_keys(d):
    r = {}
    for k, v in d:
        if k.isdigit():
            k = int(k)
        r[k] = v
    return r



    
class ModelParameters(object):
    def __init__(self, config_path=''):
        with open(config_path, 'r') as f:
                self.param = json.loads(f.read(), object_pairs_hook=int_keys)
                
        for k in ['mid_side', 'mid_side_b', 'mid_side_b2', 'stereo_w', 'stereo_n', 'reverse']:
            if not k in self.param:
                self.param[k] = False
                
        if N_BINS in self.param:
            self.param['bins'] = self.param[N_BINS]



import warnings
warnings.filterwarnings("ignore")
import os
# os.environ["CUDA_DEVICE_ORDER"]= '0'
import sys
import math
from .sad_model import SAD_INA_MODEL
import numpy as np
import torch
seed = 0
torch.manual_seed(seed)
from .thread_returning import ThreadReturning

import shutil
import time
import random
from tqdm import tqdm


from .viterbi import viterbi_decoding
from .viterbi_utils import pred2logemission, diag_trans_exp, log_trans_exp

from .features import media2feats




def _energy_activity(loge, ratio=0.4):   ##########0.9

    threshold = np.mean(loge[np.isfinite(loge)]) + np.log(ratio)
    raw_activity = (loge > threshold)
    return viterbi_decoding(pred2logemission(raw_activity),
                            log_trans_exp(150, cost0=-5))

#exp(150, cost0=-5)


def MVN(x):
   current_mean = torch.mean(x, dim=0).detach()
   current_std = torch.std(x, dim=0).detach()
   x= (x - current_mean) / current_std
   return (x) 
   
def vaw_torch(arr_in , window_shape , step):
      ndim = len(arr_in.size())
      steps = (step,) * ndim      
      arr_shape = np.array(arr_in.size())    
      window_shape = np.array(window_shape, dtype=arr_shape.dtype)
      slices = tuple(slice(None, None, st) for st in steps)         
      window_strides = np.array((arr_in.stride()))    
      indexing_strides = arr_in[slices].stride()      
      win_indices_shape = (((np.array(arr_in.size()) - np.array(window_shape))// np.array(steps)) + 1)    
      new_shape = tuple(list(win_indices_shape) + list(window_shape))      
      strides = tuple(list(indexing_strides) + list(window_strides))      
      arr_out = torch.as_strided(arr_in, new_shape, strides)
      return arr_out
   
    
def _get_patches(mspec, w, step):
      h = mspec.size(1)
      arr_out = vaw_torch(mspec, (w,h), step=step)
      arr_out = arr_out.view (arr_out.size(0), w*h) 
      arr_out =( arr_out-  torch.mean(arr_out, axis=1).view((len(arr_out), 1))) /(torch.std(arr_out,axis=1).view(arr_out.size(0),-1) + 1e-10)
      lfill = [arr_out[0,:].view(1, h*w)] * (w // (2 * step))
      rfill = [arr_out[-1,:].view(1, h*w)] * (w // (2* step) - 1 + mspec.size(0) % 2)
      arr_out=torch.vstack(lfill + [arr_out] + rfill )
      arr_out=arr_out.view(arr_out.size(0),w,h  )
      return arr_out


def _binidx2seglist(binidx):
    """
    ss._binidx2seglist((['f'] * 5) + (['bbb'] * 10) + ['v'] * 5)
    Out: [('f', 0, 5), ('bbb', 5, 15), ('v', 15, 20)]
    
    #TODO: is there a pandas alternative??
    """
    curlabel = None
    bseg = -1
    ret = []
    for i, e in enumerate(binidx):
        if e != curlabel:
            if curlabel is not None:
                ret.append((curlabel, bseg, i))
            curlabel = e
            bseg = i
    ret.append((curlabel, bseg, i + 1))
    return ret


class DnnSegmenter:
    """
    DnnSegmenter is an abstract class allowing to perform Dnn-based
    segmentation using Keras serialized models using 24 mel spectrogram
    features obtained with SIDEKIT framework.

    Child classes MUST define the following class attributes:
    * nmel: the number of mel bands to used (max: 24)
    * viterbi_arg: the argument to be used with viterbi post-processing
    * model_fname: the filename of the serialized keras model to be used
        the model should be stored in the current directory
    * inlabel: only segments with label name inlabel will be analyzed.
        other labels will stay unchanged
    * outlabels: the labels associated the output of neural network models
    """
    def __init__(self, batch_size, vad_type , model_path,tq, device):
        # load the DNN model
      if (vad_type!='vad'):
        xx = SAD_INA_MODEL()   

        
        sample_input = [torch.rand(16,68,21)]
        
        xx.load_state_dict(torch.load(model_path , map_location=torch.device(device)))  
        
        xx.eval()
        
        if (device!="cuda"):
          xx = torch.jit.trace(xx, sample_input)

          xx = torch.jit.freeze(xx)

        print('model Loded from--> ' + model_path)
        self.nn=xx.to(device)
        self.nn.eval()    
        # self.nn.summary()
        self.batch_size = batch_size
        self.tq= tq
        self.device= device
        
    def __call__(self, mspec, lseg):
        """
        *** input
        * mspec: mel spectrogram
        * lseg: list of tuples (label, start, stop) corresponding to previous segmentations
        * difflen: 0 if the original length of the mel spectrogram is >= 68
                otherwise it is set to 68 - length(mspec)
        *** output
        a list of adjacent tuples (label, start, stop)
        """

        mspec = torch.squeeze(mspec)
        
        mspec =mspec[:,0:21].clone()
        
        patches = _get_patches(mspec, 68, 2)

        batch = []
        for lab, start, stop in lseg:
            if lab == self.inlabel:
                batch.append(patches[start:stop, :])
        
        with torch.no_grad():
         bsize=self.batch_size
         if len(batch) > 0:
            batch = torch.cat(batch,0).to(self.device)
            fsize = len(batch)
            num = (fsize//bsize)+1
            x=[]
            with torch.inference_mode():
              if (self.tq == 1):
               for i in tqdm(range(num)):
                 inp =  batch[i*bsize:(i+1)*bsize,:,:]
                 if (len(inp)>0):
                    rawpred = self.nn(inp)
                    x.append(rawpred)
               rawpred = torch.concat(x)
              else:
               for i in range(num):
                 inp =  batch[i*bsize:(i+1)*bsize,:,:]
                 if (len(inp)>0):
                    rawpred = self.nn(inp)
                    x.append(rawpred)
               rawpred = torch.concat(x)              

            rawpred=rawpred.cpu().detach().numpy()
        
        ret = []
        for lab, start, stop in lseg:
            if lab != self.inlabel:
                ret.append((lab, start, stop))
                continue

            l = stop - start
            r = rawpred[:l] 
            rawpred = rawpred[l:]
           # r[finite[start:stop] == False, :] = 0.5
            pred = viterbi_decoding(np.log(r), diag_trans_exp(self.viterbi_arg, len(self.outlabels)))
            for lab2, start2, stop2 in _binidx2seglist(pred):
                ret.append((self.outlabels[int(lab2)], start2+start, stop2+start))            
        return  ret





class SpeechMusicNoise(DnnSegmenter):
    # Voice activity detection: requires energetic activity detection
    outlabels = ('speech', 'noise', 'music')
    inlabel = 'energy'
    nmel = 21
    viterbi_arg = 80

    

def getnew(ar, _len):
    
    aa = ar.numpy()
    while(len(aa)<_len):
        
      di = _len - len(aa)
      
      a2 = ar[0:di]
      
      aa = np.vstack((aa , a2))
      
    return torch.tensor(aa)



class Segmenter:


    def __init__(self, vad_type = 'sad', sr=8000, batch_size=32 ,model_path=r"C:\sad_model_pytorch.pth" , max_time=120 ,tq=1,ffmpeg_path='c:\\ffmpeg.exe', device='cuda' , pad=False):
     
        self.sample_rate = sr
        self.device = device
       
        self.max_time = max_time #sec
        self.tq=tq
        self.ffmpeg_path=ffmpeg_path
        self.pad=pad
        


        self.vad = SpeechMusicNoise(batch_size , vad_type,model_path,tq , device)

       
        self.vad_type = vad_type
        self.model_path = model_path

    def segment_feats(self, mspec, loge, start_sec):
        """
        do segmentation
        require input corresponding to wav file sampled at 16000Hz
        with a single channel
        """

        speech_len=0


        # perform energy-based activity detection
        lseg = []
        vadseg=[]
        for lab, start, stop in _binidx2seglist(_energy_activity(loge)[::2]):
            if lab == 0:
                lab = 'noEnergy'
            else:
                lab = 'energy'
                speech_len = speech_len + stop  - start
                vadseg.append(('speech', start, stop))
            lseg.append((lab, start, stop))
        if (self.vad_type == 'vad'):
           return  speech_len*0.02 , [(lab, start_sec + start * .02, start_sec + stop * .02 , stop-start) for lab, start, stop in vadseg]
        # perform voice activity detection
        lseg = self.vad(mspec, lseg)
        speech_len=0
        for lab, start, stop in lseg :
           if (lab=='speech'):
               speech_len = speech_len+ stop - start
        
        # perform gender segmentation on speech segments
      
     
        return   speech_len*0.02 , [[lab, start_sec + start * .02, start_sec + stop * .02 , (stop-start) * .02] for lab, start, stop in lseg if (lab=='speech')]   


    def __call__(self, medianame, start_sec=None, stop_sec=None):
        
        
        hop = int((10 / 1000) * self.sample_rate) # hope len = 10 ms
        
        max_frame = int(self.max_time * self.sample_rate / hop)
             
        need_speech = 2*self.max_time
        startfrom=0
        
        last_mfcc=[]
        last_isig=[]
        spee =0
        total_len=0
        while(True):
             
              mspec,loge = media2feats(medianame,self.sample_rate, startfrom , need_speech, ffmpeg_path=self.ffmpeg_path)
              
              if (mspec==[]):
                 break

              if start_sec is None:
                  start_sec = 0
              
              speech_len , isii = self.segment_feats(mspec, loge, start_sec)
             
              spee = spee + speech_len
              
              if (spee >=  self.max_time):
                  p = just_filter(isii , mspec.squeeze() , sr = self.sample_rate )
                  if (p!=[]):
                    last_mfcc.append(p)
                    total_len = total_len + (p.size()[0])
                  return torch.cat(last_mfcc)[0:max_frame,:] , self.max_time
              else:
                  p=just_filter(isii , mspec.squeeze(),sr=self.sample_rate)
                  if (p!=[]):
                    last_mfcc.append(p)
                    total_len = total_len + (p.size()[0])
                  startfrom = startfrom +  need_speech
                

        if (len(last_mfcc)==0):
           return []
        
        if total_len >=max_frame:
           return torch.cat(last_mfcc)[0:max_frame,:] , self.max_time
        else:
          if (self.pad):
             return getnew(torch.cat(last_mfcc) , max_frame) , (total_len*hop)/self.sample_rate
          else:
              return torch.cat(last_mfcc)  , (total_len*hop)/self.sample_rate         

    

def just_filter(isig , mfccs , sr=8000):   #max_time in second

   if (len(isig)==0):
     return []   
   
   # filter_mfcc_smaller_than = 10 # smaller 10 frame ==  100msec 
     
   hop = int((10 / 1000) * sr) # hope len = 10 ms
   
   nmfccs = []   
   for si in isig:
   
      t0 = int(si[1] * sr / hop)
      t1 = int(si[2] * sr / hop)
      
      # if (t1-t0 <= filter_mfcc_smaller_than ):
            # continue
      
      nmfccs.append( mfccs[t0:t1])
    
   
   if (len(nmfccs) == 0):
        return []   
        
   return (torch.cat(nmfccs))   


import requests
url = 'http://127.0.0.1:5000/test'
data = '{  "filename": "iman"}'
response = requests.post(url, data=data,headers={"Content-Type": "application/json"})
print(response.text)
print(response.json())


from iman import *
import re
from  iman.num2fa import words

rep_nothing=['ّ' , 'ِ','َ','ُ','ء','ـ','ً']
rep_space=['!','،','؟','؛','–',',',';','	' , '(',')', '[',']', '{','}', '~','!', '@','#', '$','%', '^','&','*','>','<','»','«','/',':','...'  , '.']


def conv_horoof(_str):
    t = _str.split(':')
    t0 = words(int(t[0].strip()))
    t1 = words(int(t[1].strip()))
    return " " + t0 + " و " + t1 + " دقیقه "

def cr(_str):
   return _str.replace('ي','ی').replace('ك','ک').replace('ى','ی').replace('إ','ا').replace('أ','ا').replace('ؤ','و')
   
def cf(_str):
      _str=cr(_str)
      for g in rep_nothing:
            _str = _str.replace(g , '')
      for g in rep_space:
            _str = _str.replace(g , ' ')   
      return (_str)      

def cn(text):

     pattern=r'[0-9]+[:][0-9]+'        
     x = re.search(pattern, text) 
     while(x!=None):
       y = text[x.span(0)[0]:x.span(0)[1]]
       yy = conv_horoof(y)
       text = text[:x.span(0)[0]] + yy + text[x.span(0)[1]:] 
       x = re.search(pattern, text) 
       
     pattern=r'[+-]?([0-9]*[.])?[0-9]+'    
     x = re.search(pattern, text) 
     while(x!=None):
       y = text[x.span(0)[0]:x.span(0)[1]]
       print(y)
       yy = " "  + words(y) + " "
       text = text[:x.span(0)[0]] + yy + text[x.span(0)[1]:] 
       x = re.search(pattern, text)  

    

     return text       
  

def getd(replist):
   mwords=[]
   words=[]
   ph=[]

   with open(replist , encoding='utf-8') as fr:
      for line in fr:
          t = cr(line).split('\t')
          mwords.append(t[0])
          words.append(t[1])
          ph.append(t[2].strip())
   return(mwords,words,ph)       
   
   
class normal():
  def __init__(self,replist='c:\\Replace_List.txt'):
    self.replist=replist
    self.mwords,self.words,self.ph = getd(replist) 
    
  def rep(self , _str):
     _str= cf(cn(_str))
     for i in range(len(self.words)):
       if (self.words[i] in _str):
           pattern = "(^|\s)" + self.words[i] + "($|\s)"
           _str = re.sub(pattern ," " + self.mwords[i] + " " ,  _str)
       if (self.ph[i] in _str):
           pattern = "(^|\s)" + self.ph[i] + "($|\s)"
           _str = re.sub(pattern ," " + self.mwords[i] + " " ,  _str)
                 
     return(_str.strip())  

  def from_file(self , filename ,file_out=None):

   matn  = Read(filename)

   if (file_out==None):
       file_out= PN(filename) + '_normal' + PE(filename)
   fid = open(file_out , 'w',encoding='utf-8')
   fid.write(self.rep(matn)) 
   fid.close()    




# Copyright 2021-present Kensho Technologies, LLC.
from .alphabet import Alphabet  # noqa
from .decoder import BeamSearchDecoderCTC, build_ctcdecoder  # noqa
from .language_model import LanguageModel  # noqa


__package_name__ = "pyctcdecode"
__version__ = "0.1.1"



import glob
import threading

def typethis(a):
   print (a)

threads=[]
for i in range(10):
   t = threading.Thread(target=typethis , args=(i ))
   t.start()
   threads.append(t)
   
for t in threads:
   t.join()  #wait to execute all threads

print('done')




import tensorflow as tf
from tensorflow.python.framework.convert_to_constants import convert_variables_to_constants_v2

def flops(model):
    # Convert the model
    full_model = tf.function(lambda x: model(x))
    full_model = full_model.get_concrete_function(
        tf.TensorSpec(model.inputs[0].shape, model.inputs[0].dtype))

    # Get the frozen ConcreteFunction
    frozen_func = convert_variables_to_constants_v2(full_model)
    frozen_func.graph.as_graph_def()

    # Calculate FLOPs with tf.profiler
    run_meta = tf.compat.v1.RunMetadata()
    opts = tf.compat.v1.profiler.ProfileOptionBuilder.float_operation()
    
    flops = tf.compat.v1.profiler.profile(graph=frozen_func.graph,
                                          run_meta=run_meta, cmd='op', options=opts)
    
    return flops.total_float_ops
    
    
def paramp(model):
    from prettytable import PrettyTable
    table = PrettyTable()
    table.field_names = ["Layer Type", "Output Shape", "Param #"]
    for layer in model.layers:
        table.add_row([layer.__class__.__name__, layer.output_shape, layer.count_params()])
    print(table)    
    return(model.count_params())    
    
def param(model):
    return(model.count_params())        
    
    
def gpu():
    x = tf.test.is_gpu_available()
    print(x)
    return(x)


def gpun():
    gpus =   tf.config.list_physical_devices('GPU')
    try:
      num_of_gpus = len(gpus)
      print('Number of GPUs Available: ', num_of_gpus)    
      return num_of_gpus
    except:
         print('No GPU')
         return 0         
    
    
def limit():  #only allocate as much GPU memory based on runtime allocations
     gpus = tf.config.experimental.list_physical_devices('GPU')
     if gpus:
       try:
          for gpu in gpus:
            tf.config.experimental.set_memory_growth(gpu, True)
       except RuntimeError as e:
            print(e)


from flask import logging, Flask, render_template, request, flash,send_from_directory,jsonify

app = Flask(__name__)


@app.route('/get', methods=['get'])
def hello_world():
    return jsonify({'message' : 'Hello, World!'})

@app.route('/test', methods=['post'])
def test():
    x= request.get_json()
    filename=x['filename']
    return jsonify({'message' : filename})
 
    
if __name__ == "__main__":
    # app.run(host='192.168.43.37' ,port='8080', debug=True)
    app.run(debug=True)



from torch import nn


def param(model):
    return(sum(p.numel() for p in model.parameters()) , sum(p.numel() for p in model.parameters() if p.requires_grad))
    
    
def paramp(model):
    from prettytable import PrettyTable
    table = PrettyTable(["Modules", "Parameters"])
    total_params = 0
    for name, parameter in model.named_parameters():
        if not parameter.requires_grad:
            continue
        params = parameter.numel()
        table.add_row([name, params])
        total_params += params
    print(table)
    print(f"Total Trainable Params: {total_params}")
    return total_params

def layers(model):
    children = list(model.children())
    return [model] if len(children) == 0 else [ci for c in children for ci in layers(c)]
    

def gpu():
    import torch
    x = torch.cuda.is_available()  
    print(x)
    return(x)

def gpun():
    import torch
    num_of_gpus =   torch.cuda.device_count()  
    print('Number of GPUs Available: ', num_of_gpus)    
    for i in range(num_of_gpus):
        print('GPU ', i, ' : ', torch.cuda.get_device_name(i))
    return num_of_gpus



import numpy as np
from scipy import signal
import os
import scipy.io.wavfile as wave
from numpy.lib.stride_tricks import as_strided
import six
import subprocess
import platform
from iman import *

_system = platform.system().lower()



def Read_Alaw_1(filename):
    with open(filename, "rb") as binaryfile :
       myArr = bytearray(binaryfile.read())
    g=[ALawDecompressTable[x]/32768 for x in myArr]
    return np.array(g , dtype=np.float32)
    
    
def Read_Alaw(filename,sr=8000,start_from=0,dur=-1,ffmpeg_path='c:\\ffmpeg.exe'):

           ffmpeg_command = [ffmpeg_path, "-ss" , str(start_from) ,  "-ar","8000", "-f" ,"alaw" ,'-i', filename, '-ac', "1", '-ar', str(sr), '-f', 'wav', 'pipe:1']
           if (dur!=-1):
               ffmpeg_command = [ffmpeg_path, "-ss" , str(start_from) , "-t" ,str(dur) , "-ar","8000", "-f" ,"alaw" ,'-i', filename, '-ac', "1", '-ar', str(sr), '-f', 'wav', 'pipe:1']            
           
           pipe = subprocess.run(ffmpeg_command, stdout=subprocess.PIPE,   stderr=subprocess.PIPE,  bufsize=10**8)
           
           audio_np = np.frombuffer(buffer=pipe.stdout, dtype=np.uint16, offset=8*44)
             
           x =np.where(audio_np>32768)    [0]    
           y =np.where(audio_np<=32768)    [0]  
           
           out_wav = np.zeros(len(audio_np) , dtype=np.float32)
           out_wav[x] = (audio_np[x] - 32768)/32768-1
           out_wav[y] = audio_np[y] /32768
           
          
           return(out_wav) 
       

def Resample(data , fs, sr):
    """Return Resampled Data.

    Parameters
    ----------
    data : input audio data
    fs : data sampling rate
    sr : resample data to sr
    
    Output
    ----------
    data with sampling rate --> sr

    """
    if (fs!=sr):
        nesbat = sr/fs
        m = np.max(np.abs(data))
        data = 0.5*(data/(m+1e-6))
        data = signal.resample(data, int(len(data)*nesbat))
    return (np.array(data , dtype=np.float32))    

def ReadMp3_miniaudio(filename,sr,mono=True):
    import miniaudio
    mp3file = miniaudio.read_file(filename,True)
    sample_rate = mp3file.sample_rate
    data =  np.array(mp3file.samples , dtype=np.float32)/32768
    if (mp3file.nchannels==1):
       if (sample_rate!=sr):
            data = Resample(data ,sample_rate ,sr )
    else:
         ch1=np.array( data[::2] , dtype=np.float32)
         ch2=np.array(data[1::2], dtype=np.float32)
         if (mono):
           cha = ch1+ch2
         if (sample_rate!=sr):
                if (mono):
                  cha = Resample(cha ,sample_rate ,sr )
                else:  
                   ch1 = Resample(ch1 ,sample_rate ,sr )
                   ch2 = Resample(ch2 ,sample_rate ,sr )
         if (mono):          
           data = cha
         else:
           data=  [ch1,ch2]       
    return(data)




def ReadFFMPEG(filename,sr=16000,start_from=0,dur=-1,mono = False , ffmpeg_path='c:\\ffmpeg.exe',ffprobe_path='c:\\ffprobe.exe'):
           
           chnumber = 1
           if (not mono):
             ffprobe_command = [ffprobe_path ,  "-v","error","-show_entries","stream=channels","-of","default=nw=1" ,filename ] 
             pipe1 = subprocess.run(ffprobe_command, stdout=subprocess.PIPE)
             chnumber = int(str(pipe1.stdout).strip().split('=')[1].split('\\')[0])
             
           
           
           ffmpeg_command = [ffmpeg_path, '-ss' , str(start_from), '-i', filename,  '-ac' ,str(chnumber) , '-ar', str(sr), '-f', 'wav', 'pipe:1']
           
           if (dur!=-1):
                    ffmpeg_command = [ffmpeg_path, '-ss' , str(start_from), '-t' , str(dur) , '-i', filename, '-ac', str(chnumber), '-ar', str(sr), '-f', 'wav', 'pipe:1']
           
           pipe = subprocess.run(ffmpeg_command, stdout=subprocess.PIPE,   stderr=subprocess.PIPE,  bufsize=10**8)
           
           audio_np = np.frombuffer(buffer=pipe.stdout, dtype=np.uint16, offset=8*44)
             
           x =np.where(audio_np>32768)    [0]    
           y =np.where(audio_np<=32768)    [0]  
           
           out_wav = np.zeros(len(audio_np) , dtype=np.float32)
           out_wav[x] = (audio_np[x] - 32768)/32768-1
           out_wav[y] = audio_np[y] /32768
           
           
           x=[]
           if (chnumber==1):
              x.append(out_wav)
              if (mono):
                 return x[0]
              else:   
                 return(x) 
           else:
              x.append(out_wav[1::2])
              x.append(out_wav[::2]) 
              return(x)



def Read(filename,sr=8000, start_from=0, dur=-1,mono = False,ffmpeg_path='c:\\ffmpeg.exe',ffprobe_path='c:\\ffprobe.exe' ):   # output sampling rate is sr
   """Return Audio Data. (Just Mono Files)

    Parameters
    ----------
    filename : input file path (PCM or alaw or [alaw_raw with .l or .r ext] or mp3)
    sr : desired sampling rate
    
    Output
    ----------
    data with sampling rate --> sr

   """   
   ext = os.path.basename(filename).split('.')[-1] 
   if (ext.lower() == 'l' or ext.lower() == 'r' or os.path.basename(filename).lower().endswith('.r.wav') or os.path.basename(filename).lower().endswith('.l.wav') or os.path.basename(filename).lower().endswith('.rtemp.wav') or os.path.basename(filename).lower().endswith('.ltemp.wav')):
         return( Read_Alaw(filename,sr=sr ,ffmpeg_path=ffmpeg_path ,start_from=start_from,dur=dur))
       
  
   return(ReadFFMPEG(filename,sr=sr, start_from=start_from,dur=dur,mono =mono ,  ffmpeg_path=ffmpeg_path,ffprobe_path=ffprobe_path ))
    
   
   
   # if (ext.lower() == 'mp3'):
         # return(ReadMp3(filename,sr=sr, mono=True,ffmpeg_path=ffmpeg_path))
   
   # File_Type_Header= np.fromfile(filename, dtype=np.byte, count=4, offset=8)
   # File_Type_Header = "".join(map(chr, File_Type_Header))   # Must be WAVE
   # if (File_Type_Header!="WAVE"):
       # return(ReadMp3(filename,sr=sr, mono=True,ffmpeg_path=ffmpeg_path))
       
   # Type_of_format = np.fromfile(filename, dtype=np.int8, count=1, offset=20)[0]  # 1 is PCM   6 is alaw

   # ch = np.fromfile(filename, dtype=np.int8, count=1, offset=22)[0]  
   
   # if (ch!=1):
       # return(ReadMp3(filename,sr=sr, mono=True,ffmpeg_path=ffmpeg_path))
   
   # fs = np.fromfile(filename, dtype=np.int32, count=1, offset=24)[0] # Hz
   
   # if (Type_of_format!=1 and Type_of_format!=6 ):
       # return(ReadMp3(filename,sr=sr, mono=True,ffmpeg_path=ffmpeg_path))
   
   # if (Type_of_format==6):   # if file is alaw
        # byte_length = np.fromfile(filename, dtype=np.int32, count=1, offset=40)[0]
        # data = np.fromfile(filename, dtype=np.byte, count=byte_length*4, offset=44)
        # data=np.array([ALawDecompressTable[x]/32768 for x in data], dtype=np.float32)
        # return (Resample(data,fs,sr))
   
   # chunk_header = np.fromfile(filename, dtype=np.byte, count=4, offset=36)
   # chunk_header = "".join(map(chr, chunk_header))
   # if (chunk_header!="data"):
     # if (_system == 'windows'):
      # data=ReadMp3(filename,sr=sr, mono=True,ffmpeg_path=ffmpeg_path)
     # else:
      # import librosa
      # data,_ = librosa.load(filename , sr)      
   # elif(chunk_header=="data"):
      # byte_length = np.fromfile(filename, dtype=np.int32, count=1, offset=40)[0]
      # data = np.fromfile(filename, dtype=np.int16, count=byte_length // 2, offset=44)/32768
      # data = Resample(data,fs,sr)
   # return (data)

def Write(filename, data ,fs):

   """Write Audio Data.

    Parameters
    ----------
    filename : Output file path (.wav)
    data : Audio data
    fs : sampling rate
    
    Output
    ----------
    Nothing

   """
   wave.write(filename, fs, np.int16(data*32768))
   
def WriteS(filename, data ,fs):

   """Write Audio Data Sterio.

    Parameters
    ----------
    filename : Output file path (.wav)
    data : Audio data
    fs : sampling rate
    
    Output
    ----------
    Nothing

   """
   xx = np.asarray( [np.int16(data*32768) , np.int16(data*32768)]).transpose()

   wave.write(filename, fs, xx)
def rmse(y=None, S=None, frame_length=2048, hop_length=512,
         center=True, pad_mode='reflect'):
    
    if y is not None and S is not None:
        raise ValueError('Either `y` or `S` should be input.')
    if y is not None:
        y = (y)
        if center:
            y = np.pad(y, int(frame_length // 2), mode=pad_mode)

        x = frame(y,
                       frame_length=frame_length,
                       hop_length=hop_length)
    elif S is not None:
        x, _ = _spectrogram(y=y, S=S,
                            n_fft=frame_length,
                            hop_length=hop_length)
    else:
        raise ValueError('Either `y` or `S` must be input.')
    return np.sqrt(np.mean(np.abs(x)**2, axis=0, keepdims=True))
def frame(y, frame_length=2048, hop_length=512):
   
    if not isinstance(y, np.ndarray):
        raise ParameterError('Input must be of type numpy.ndarray, '
                             'given type(y)={}'.format(type(y)))

    if y.ndim != 1:
        raise ParameterError('Input must be one-dimensional, '
                             'given y.ndim={}'.format(y.ndim))

    if len(y) < frame_length:
        raise ParameterError('Buffer is too short (n={:d})'
                             ' for frame_length={:d}'.format(len(y), frame_length))

    if hop_length < 1:
        raise ParameterError('Invalid hop_length: {:d}'.format(hop_length))

    if not y.flags['C_CONTIGUOUS']:
        raise ParameterError('Input buffer must be contiguous.')

    # Compute the number of frames that will fit. The end may get truncated.
    n_frames = 1 + int((len(y) - frame_length) / hop_length)

    # Vertical stride is one sample
    # Horizontal stride is `hop_length` samples
    y_frames = as_strided(y, shape=(frame_length, n_frames),
                          strides=(y.itemsize, hop_length * y.itemsize))
    return y_frames	
def power_to_db(S, ref=1.0, amin=1e-10, top_db=80.0):
   

    S = np.asarray(S)

    if amin <= 0:
        raise ParameterError('amin must be strictly positive')

    if np.issubdtype(S.dtype, np.complexfloating):
        warnings.warn('power_to_db was called on complex input so phase '
                      'information will be discarded. To suppress this warning, '
                      'call power_to_db(magphase(D, power=2)[0]) instead.')
        magnitude = np.abs(S)
    else:
        magnitude = S

    if six.callable(ref):
        # User supplied a function to calculate reference power
        ref_value = ref(magnitude)
    else:
        ref_value = np.abs(ref)

    log_spec = 10.0 * np.log10(np.maximum(amin, magnitude))
    log_spec -= 10.0 * np.log10(np.maximum(amin, ref_value))

    if top_db is not None:
        if top_db < 0:
            raise ParameterError('top_db must be non-negative')
        log_spec = np.maximum(log_spec, log_spec.max() - top_db)

    return log_spec
def _signal_to_frame_nonsilent(y_mono, frame_length=2048, hop_length=512, top_db=60,
                               ref=np.max):

    # Convert to mono
   

    # Compute the MSE for the signal
    mse = rmse(y=y_mono,
                       frame_length=frame_length,
                       hop_length=hop_length)**2

    return (power_to_db(mse.squeeze(),
                             ref=ref,
                             top_db=None) > - top_db)					 
def split(y, top_db=40, ref=np.max, frame_length=200, hop_length=80):

    non_silent = _signal_to_frame_nonsilent(y,
                                            frame_length=frame_length,
                                            hop_length=hop_length,
                                            ref=ref,
                                            top_db=top_db)

    edges = np.flatnonzero(np.diff(non_silent.astype(int)))

    # Pad back the sample lost in the diff
    edges = [edges + 1]

    # If the first frame had high energy, count it
    if non_silent[0]:
        edges.insert(0, [0])

    # Likewise for the last frame
    if non_silent[-1]:
        edges.append([len(non_silent)])

    # Convert from frames to samples
    edges = frames_to_samples(np.concatenate(edges),
                                   hop_length=hop_length)

    # Clip to the signal duration
    edges = np.minimum(edges, y.shape[-1])

    # Stack the results back as an ndarray
    return edges.reshape((-1, 2))
    
def frames_to_samples(frames, hop_length=512, n_fft=None):
 
    offset = 0
    if n_fft is not None:
        offset = int(n_fft // 2)

    return (np.asanyarray(frames) * hop_length + offset).astype(int)
    
def ReadT(filename, sr=8000 , mono=True): 
    import torchaudio
    
    wavs, fs = torchaudio.load(filename)
    
    if (fs!=sr):
       import torchaudio.transforms as T  
       resampler = T.Resample(fs, sr, dtype=wavs.dtype)
       wavs  = resampler(wavs)
    
    if (len(wavs)==2 and mono==True):
        return ( (wavs[0] +wavs[1]).unsqueeze(0))
        
    return wavs
    
def VAD(wav,top_db=40, frame_length=200, hop_length=80):
    a = split(wav,top_db=top_db, frame_length=frame_length, hop_length=hop_length)
    v_wav=[]
    for x in a:
        v_wav.append(wav[x[0]:x[1]])
    v = np.concatenate(v_wav)
    return(v)       
    
def change_format(fname , sr=16000 , ext='mp3' , mono=True ,ffmpeg_path='c:\\ffmpeg.exe' ,oname=None ):
    chnum=2
    if (mono):
      chnum=1
      
    forext = []
    if (PE(fname).lower()=='.l' or PE(fname).lower()=='.r') :
         forext = ['-ar', '8000', '-f' ,'alaw']    
      
      
    if (not ext.startswith('.')):
       ext = '.' + ext    

    if (oname==None):
       outfile =  PJ(PD(fname) , PN(fname)+ ext)
    else:
      if ('.' in oname):
        outfile =  oname
      else:
        outfile =  oname + ext
    
    if (PX(outfile)):
       print('File Exists--> ' + PB(outfile))
       return 0
       
    try:
    
      if (forext==[]):
         ffmpeg_command = [ffmpeg_path, '-i', fname, '-ac', str(chnum), '-ar', str(sr), 'outfile']
      else:
          ffmpeg_command = [ffmpeg_path]
          ffmpeg_command.extend(forext)
          ffmpeg_command.extend(['-i', fname, '-ac', str(chnum), '-ar', str(sr), 'outfile'])

      x = cmd(ffmpeg_command )
      if (PX(outfile) and PS(outfile)>1):
         return 1
      else:
          try:
             os.remove(outfile)
          except:
             pass 
          return 0             
          
    except:
      print('ffmpeg has Problem with --> ' + PB(outfile)) 
      try:
         os.remove(outfile)
      except:
         pass
      
      return 0
 

def fun(fname , b):
   change_format(fname , sr=b[0] , ext=b[1] , mono=b[2] , ffmpeg_path=b[3] , oname=PJ(b[4] , PN(fname) + b[1] ) )
 
def compress (fname_pattern , sr=16000 , ext='mp3' , mono=True ,ffmpeg_path='c:\\ffmpeg.exe' , ofolder=None , worker=4): 
    
    if (not ext.startswith('.')):
       ext = '.' + ext    
       
    files = gf(fname_pattern) 
    
    
    if (len(files)==0):
        print('no File!!!')
        return
    
   
    if (ofolder==None):
        pp = input('It is better to set ofolder\n press y to continue compression: ')
        if (pp.lower()=='y'):
          ofolder=""
        else:
           return        
    else:
       PM(ofolder)
    
    worker=min(len(files) , worker)     

    from multiprocessing import Pool
    from functools import partial  
    with Pool(processes=worker) as pool:
         pool.map(partial(fun, b=[sr , ext,mono , ffmpeg_path,ofolder]), files)     



def clip_value(wav):

    # wav must be output of sad
    wav = abs( wav / abs(wav).max())

    cli= np.where(wav>=0.99)[0]
    cli1 = np.insert(np.delete(cli-1 ,0) , len(cli)-1,0)
    cli2=np.delete(np.insert(cli+1,0,0) , len(cli))   
    
    a=(cli == cli1 ) 
    b=(cli == cli2)
    c = np.logical_or(a, b)
    
    x = len(np.where(c==True)[0])
        
    clip_percentage = 100*x / len(wav)
    return (clip_percentage)         
    
    

ALawDecompressTable =[

     -5504, -5248, -6016, -5760, -4480, -4224, -4992, -4736,

     -7552, -7296, -8064, -7808, -6528, -6272, -7040, -6784,

     -2752, -2624, -3008, -2880, -2240, -2112, -2496, -2368,

     -3776, -3648, -4032, -3904, -3264, -3136, -3520, -3392,

     -22016,-20992,-24064,-23040,-17920,-16896,-19968,-18944,

     -30208,-29184,-32256,-31232,-26112,-25088,-28160,-27136,

     -11008,-10496,-12032,-11520,-8960, -8448, -9984, -9472,

     -15104,-14592,-16128,-15616,-13056,-12544,-14080,-13568,

     -344,  -328,  -376,  -360,  -280,  -264,  -312,  -296,

     -472,  -456,  -504,  -488,  -408,  -392,  -440,  -424,

     -88,   -72,   -120,  -104,  -24,   -8,    -56,   -40,

     -216,  -200,  -248,  -232,  -152,  -136,  -184,  -168,

     -1376, -1312, -1504, -1440, -1120, -1056, -1248, -1184,

     -1888, -1824, -2016, -1952, -1632, -1568, -1760, -1696,

     -688,  -656,  -752,  -720,  -560,  -528,  -624,  -592,

     -944,  -912,  -1008, -976,  -816,  -784,  -880,  -848,

      5504,  5248,  6016,  5760,  4480,  4224,  4992,  4736,

      7552,  7296,  8064,  7808,  6528,  6272,  7040,  6784,

      2752,  2624,  3008,  2880,  2240,  2112,  2496,  2368,

      3776,  3648,  4032,  3904,  3264,  3136,  3520,  3392,

      22016, 20992, 24064, 23040, 17920, 16896, 19968, 18944,

      30208, 29184, 32256, 31232, 26112, 25088, 28160, 27136,

      11008, 10496, 12032, 11520, 8960,  8448,  9984,  9472,

      15104, 14592, 16128, 15616, 13056, 12544, 14080, 13568,

      344,   328,   376,   360,   280,   264,   312,   296,

      472,   456,   504,   488,   408,   392,   440,   424,

      88,    72,   120,   104,    24,     8,    56,    40,

      216,   200,   248,   232,   152,   136,   184,   168,

      1376,  1312,  1504,  1440,  1120,  1056,  1248,  1184,

      1888,  1824,  2016,  1952,  1632,  1568,  1760,  1696,

      688,   656,   752,   720,   560,   528,   624,   592,

      944,   912,  1008,   976,   816,   784,   880,   848]


import math


# default parameters for decoding (can be modified)
DEFAULT_ALPHA = 0.5
DEFAULT_BETA = 1.5
DEFAULT_UNK_LOGP_OFFSET = -10.0
DEFAULT_BEAM_WIDTH = 100
DEFAULT_HOTWORD_WEIGHT = 10.0
DEFAULT_PRUNE_LOGP = -10.0
DEFAULT_PRUNE_BEAMS = False
DEFAULT_MIN_TOKEN_LOGP = -5.0
DEFAULT_SCORE_LM_BOUNDARY = True

# other constants for decoding
AVG_TOKEN_LEN = 6  # average number of characters expected per token (used for UNK scoring)
MIN_TOKEN_CLIP_P = 1e-15  # clipping to avoid underflow in case of malformed logit input
LOG_BASE_CHANGE_FACTOR = 1.0 / math.log10(math.e)  # kenlm returns base10 but we like natural



from iman import *
import datetime
import pytz
from win32com.propsys import propsys, pscon


def get_age(a):
    years = a//365
    rooz_baqimande= a%365
    
    mah = rooz_baqimande//30
    rooz_baqimande_2 = rooz_baqimande%30
    
    return ("%sy_%sm_%sd" %(D(years,2) , D(mah,2) , D(rooz_baqimande_2,2)))

def get(fname):
   properties = propsys.SHGetPropertyStoreFromParsingName(fname)
   dt = properties.GetValue(pscon.PKEY_Media_DateEncoded).GetValue()
   t = datetime.date(dt.year,dt.month,dt.day)
   return (t)


import matplotlib.pyplot as plt
import time
from glob import glob as gf
import os
import numpy as np
from joblib import load,dump


def help():
    from iman import web
    x=web.dl(r'https://pypi.org/project/iman/')
    idx = x.index('class="project-description"')
    idy = x.index('</div>' , idx+1)
    y= (x[idx+29:idy].strip().replace('<span class="docutils literal">','<').replace('</span>','>').replace('</section>','').replace('<h2>','').replace('</h2>','').replace('<p>','').replace('</p>','').replace('<section id=from-iman-import>',''))
    

    while(True):
       try:
        idx=0
        idx = y.index('<section',idx+1)

        idy = y.index('>' , idx+1)

        y = y.replace(y[idx:idy+1] , '')
       except:
         break       
    print(y)
   
def clear():
    if os.name == 'nt':
        _ = os.system('cls')  
    else:
        _ = os.system('clear')
        
def now():
   return time.time()
     
def F(float_number , float_number_count = 2):
   _str=("{:." + str(float_number_count) +"f}").format(float_number)
   return(_str) 

def D(int_number , int_number_count = 3):
   _str=("{:0>" + str(int_number_count) +"d}").format(int(int_number))
   return(_str) 
   
def Write(_str,Filename):
   with open(Filename , 'w' , encoding='utf-8') as fid:
              fid.write(_str)    

def Write_List(MyList,Filename):
   with open(Filename , 'w' , encoding='utf-8') as fid:
        for x in MyList:
              fid.write(str(x) + '\n')    

def Write_Dic(MyDic,Filename):
   with open(Filename , 'w' , encoding='utf-8') as fid:
        for x,y in MyDic.items():
              fid.write(str(x) + '\t' + str(y) + '\n')                 
              
def Read(Filename):
    with open(Filename , 'r' , encoding='utf-8') as fid:
         return(fid.read())

def Read_Lines(Filename):
    with open(Filename , 'r' , encoding='utf-8') as fid:
         return([x.strip() for x in fid if (x.strip()!="")])    

def gfa(directory , ext="*.*"):
 fols = gf(directory)
 a=[]
 for _fol in fols: 
   [a.append(x) for x in gf(os.path.join(_fol , ext))]
   for root, dirs, files in os.walk(_fol):
      for dirname in dirs:
         _dir =os.path.join(root, dirname)         
         [a.append(x) for x in gf(os.path.join(_dir , ext))]           
 return a         

def ReadE(Filename):
    import pandas as pd
    pp = pd.read_excel(Filename , engine='openpyxl')
    return pp
    
def PB(filename):
   return os.path.basename(filename)    
         
def PD(filename):
   return os.path.dirname(filename)  
   
def PM(folname):
   return os.makedirs(folname , exist_ok=True)
       
def PN(filename):
   return os.path.basename(os.path.splitext(filename)[0])  

def PE(filename):
   return os.path.splitext(filename)[1]   

def PX(filename):
   return os.path.exists(filename)    
   
def PJ(*_segments):
   x=""
   for p in _segments:
      x = os.path.join(x,p)
   return x   
   
def PS(filename):
   return os.path.getsize(filename)  
   
def RI(start_int , end_int , count=1):
   return np.random.randint(low=int(start_int), high=int(end_int), size=(count,))
   
def RF(start_float , end_float , count=1):
  return np.random.uniform(start_float, end_float , count)
  
def RS(Arr):
  np.random.shuffle(Arr)
  return Arr
  
def RC(Arr , _size=1):
   return np.random.choice(Arr , _size)


def LJ(job_file_name):
   return load(job_file_name)
   
def SJ(value , job_file_name):
   dump(value , job_file_name)   
   
def LN(np_file_name):
   return np.load(np_file_name) 
   
def SN(arr , np_file_name):
   return np.save(np_file_name , arr)   
   
def cmd(command , redirect=True):
  if (redirect):
   return os.popen(command).read()   
  else:
    os.system(command)  
    
    
def onehot(data, nb_classes):
    """Convert an iterable of indices to one-hot encoded labels."""
    targets = np.array(data).reshape(-1)
    return np.eye(nb_classes)[targets]
    
    
def exe(fname):
    cmd('pyinstaller --onefile ' + fname , False)    
    
def FWL(wavfolder , sr): #Get Audio Lenth Exist in this folder    
    files=gfa(wavfolder , '*.wav')
    n=0
    for filename in files:
         n=n+PS(filename)
    header=len(files)*44       
    print((n-header)/(sr*2*60*60))
    return  ((n-header)/(sr*2*60*60))       
    
    
def norm(vector):
   m = np.linalg.norm(vector)
   return (vector/m)   
   

def delete(pattern):
    files = gf(pattern)
    for fname in files:
       isFile = os.path.isfile(fname)
       isDirectory = os.path.isdir(fname)
       if (isFile):
           cmd('del /Q "%s"' %(fname) )  
       if (isDirectory):           
           cmd('rmdir /S/Q "%s"' %(fname) )    
       
       
def rename(fname , fout):
       cmd('move "%s" "%s"' %(fname , fout) )         
       
       
def separate(pattern,folout=None,model_path_folder=None):   #model_path_folder contain .th model of Dmucs
    files = gf(pattern)
    for fname in files:
        print(fname)
        if (folout):
            PM(folout)
            if (model_path_folder):
                cmd('python -m demucs.separate --repo "%s" -o "%s" "%s"' %(model_path_folder,folout,fname))  
            else:
                cmd('python -m demucs.separate -o "%s" "%s"' %(folout,fname))  
        else:
           if (model_path_folder):
                cmd('python -m demucs.separate --repo "%s" "%s"' %(model_path_folder,fname)) 
           else:
               cmd('python -m demucs.separate "%s"' %(fname))               


def dll(fname):
    text="from Cython.Distutils import build_ext\n"
    text+="from setuptools import Extension, setup\n"
    text+=('ext_modules = [Extension("%s", ["%s.py"])]\n' %(PN(fname),PN(fname)))
    text+="setup(\n"
    text+="    name='Test Program',\n"
    text+="    cmdclass={'build_ext': build_ext},\n"
    text+="    ext_modules=ext_modules,\n"
    text+=")\n"
    Write(text , "setup.py")
    cmd('python setup.py build_ext --inplace')
    delete('setup.py')    
    delete('build')    
    delete(PN(fname)+'.c')    
    
def get_hard_serial():
   x = cmd("vol c:",True)  
   return(x.split('\n')[1].split('is')[1].strip())   
   
def mute_mic():
   import win32api
   import win32gui
   WM_APPCOMMAND = 0x319
   APPCOMMAND_MICROPHONE_VOLUME_MUTE = 0x180000
   hwnd_active = win32gui.GetForegroundWindow()   
   win32api.SendMessage(hwnd_active, WM_APPCOMMAND, None, APPCOMMAND_MICROPHONE_VOLUME_MUTE)


import os
import numpy as np
from iman import Audio
from iman import Features
import torch


def media2feats(wavname,sr=16000,start_from=0 ,dur=-1,ffmpeg_path='c:\\ffmpeg.exe'):
    try:
       sig1 = Audio.Read(wavname ,sr,start_from , dur ,True, ffmpeg_path)
    except :
        return [],[]
    sig=sig1+1e-7
    sig = torch.unsqueeze( torch.from_numpy(sig) , dim=0) 
    try:    
      mfccs, mspec,loge  = Features.SB.Get(sig , sr , n_mels=40 , n_mfcc=24 , n_fft=200)   
    except:
      print('error in mfcc extraction')
      return [] ,[]   
    if (len(loge)<68):   
           return [] ,[]
    return  mfccs , loge     
  



from iman import *
from scipy.signal import fftconvolve
import subprocess


def cut_noise(noise , le):
      n=[]
      mle = len(noise)
      if (mle > le):
         return noise[0:le]
      else:
         kt = le//mle    
         for i in range(kt):
             n = np.hstack((n , noise))
         nn = noise[0:le - mle*kt]
         n = np.hstack((n , nn))
         return n
     
def Add_Noise( data , noise , snr=15):
        cutnoise = cut_noise(noise , len(data))
        iPn = 1/np.mean(noise*noise)
        Px = np.mean(data*data)
        Msnr = np.sqrt( 10**(-snr/10) * iPn*Px )
        data_noise = data + cutnoise*Msnr
        return data_noise
 

def Add_Reverb( data , rir):
      return fftconvolve(data,rir)[1000:len(data)+1000]
 
def mp3(fname,fout,sr_out,ratio,ffmpeg_path='c:\\ffmpeg.exe'):
   try:
    
    command='%s -i "%s"  -ar %s -f mp3 pipe: | %s -f mp3 -i pipe: -y -ar %s "%s" ' %(ffmpeg_path,fname ,ratio,ffmpeg_path,sr_out, fout) 
    os.system(command)

    return 1  
   except:
    return 0        
  

def speed(fname,fout,ratio,ffmpeg_path='c:\\ffmpeg.exe'):
   try:
    command='%s -i "%s"   -af "atempo=%s" "%s"' %(ffmpeg_path,fname ,ratio,fout )   
    os.system(command) 
    return 1
   except:
     return 0  

def volume(fname ,fout,ratio,ffmpeg_path='c:\\ffmpeg.exe'):
   try:
    command='%s -i "%s"   -af "volume=%s" "%s"' %(ffmpeg_path,fname ,ratio,fout ) 
    os.system(command)
    return 1
   except:
     return 0  
   
	             
def Add_NoiseT( data , noise , snr=15):
        data = data.squeeze()
        noise = noise.squeeze()
        cutnoise = cut_noise(noise , data.size(0))
        iPn = 1/torch.mean(noise*noise)
        Px = torch.mean(data*data)
        Msnr = torch.sqrt( 10**(-snr/10) * iPn*Px )
        data_noise = data + cutnoise*Msnr
        return data_noise

    
def Add_ReverbT( data , rir):
      data = data.squeeze()
      rir = rir.squeeze()
      return fftconvolve(data,rir)[1000:data.size(0)+1000]                 


# Copyright 2021-present Kensho Technologies, LLC.
from __future__ import division

import abc
import logging
import re
from typing import Collection, Iterable, List, Optional, Pattern, Set, Tuple, cast

import numpy as np
from pygtrie import CharTrie  # type: ignore

from .constants import (
    AVG_TOKEN_LEN,
    DEFAULT_ALPHA,
    DEFAULT_BETA,
    DEFAULT_HOTWORD_WEIGHT,
    DEFAULT_SCORE_LM_BOUNDARY,
    DEFAULT_UNK_LOGP_OFFSET,
    LOG_BASE_CHANGE_FACTOR,
)


logger = logging.getLogger(__name__)


try:
    import kenlm  # type: ignore
except ImportError:
    logger.warning(
        "kenlm python bindings are not installed. Most likely you want to install it using: "
        "pip install https://github.com/kpu/kenlm/archive/master.zip"
    )


def load_unigram_set_from_arpa(arpa_path: str) -> Set[str]:
    """Read unigrams from arpa file."""
    unigrams = set()
    with open(arpa_path) as f:
        start_1_gram = False
        for line in f:
            line = line.strip()
            if line == "\\1-grams:":
                start_1_gram = True
            elif line == "\\2-grams:":
                break
            if start_1_gram and len(line) > 0:
                parts = line.split("\t")
                if len(parts) == 3:
                    unigrams.add(parts[1])
    if len(unigrams) == 0:
        raise ValueError("No unigrams found in arpa file. Something is wrong with the file.")
    return unigrams


def _prepare_unigram_set(unigrams: Collection[str], kenlm_model: kenlm.Model) -> Set[str]:
    """Filter unigrams down to vocabulary that exists in kenlm_model."""
    if len(unigrams) < 1000:
        logger.warning(
            "Only %s unigrams passed as vocabulary. Is this small or artificial data?",
            len(unigrams),
        )
    unigram_set = set(unigrams)
    unigram_set = set([t for t in unigram_set if t in kenlm_model])
    retained_fraction = 1.0 if len(unigrams) == 0 else len(unigram_set) / len(unigrams)
    if retained_fraction < 0.1:
        logger.warning(
            "Only %s%% of unigrams in vocabulary found in kenlm model-- this might mean that your "
            "vocabulary and language model are incompatible. Is this intentional?",
            round(retained_fraction * 100, 1),
        )
    return unigram_set


def _get_empty_lm_state() -> kenlm.State:
    """Get unintialized kenlm state."""
    try:
        kenlm_state = kenlm.State()
    except ImportError:
        raise ValueError("To use a language model, you need to install kenlm.")
    return kenlm_state


class HotwordScorer:
    def __init__(
        self,
        match_ptn: Pattern[str],
        char_trie: CharTrie,
        weight: float = DEFAULT_HOTWORD_WEIGHT,
    ) -> None:
        """Scorer for hotwords if provided.

        Args:
            match_ptn: match pattern for hotword unigrams
            char_trie: trie for all hotwords to do partial matching
            weight: weight for score increase
        """
        self._match_ptn = match_ptn
        self._char_trie = char_trie
        self._weight = weight

    def __contains__(self, item: str) -> bool:
        """Contains."""
        return cast(bool, self._char_trie.has_node(item) > 0)

    def score(self, text: str) -> float:
        """Get total hotword score for input text."""
        return self._weight * len(self._match_ptn.findall(text))

    def score_partial_token(self, token: str) -> float:
        """Get total hotword score for input text."""
        if token in self:
            # find shortest unigram starting with the given partial token
            min_len = len(next(self._char_trie.iterkeys(token, shallow=True)))
            # scale score by length of unigram matched so far
            score = self._weight * len(token) / min_len
        else:
            score = 0.0
        return score

    @classmethod
    def build_scorer(
        cls, hotwords: Optional[Iterable[str]] = None, weight: float = DEFAULT_HOTWORD_WEIGHT
    ) -> "HotwordScorer":
        """Use hotword list to create regex pattern and character trie for scoring."""
        # make sure we get an iterable
        hotwords = hotwords or []
        # remove whitespace
        hotwords = [s.strip() for s in hotwords if len(s.strip()) > 0]
        if len(hotwords) > 0:
            hotword_unigrams = []
            for ngram in hotwords:
                # split ngrams to get words
                for unigram in ngram.split():
                    hotword_unigrams.append(unigram)

            # create pattern to match full words
            # sort by length to get longest possible match
            # use lookahead and lookbehind to match on word boundary instead of '\b' to only match
            # on space or bos/eos
            match_ptn = re.compile(
                r"|".join(
                    [
                        r"(?<!\S)" + re.escape(s) + r"(?!\S)"
                        for s in sorted(hotword_unigrams, key=len, reverse=True)
                    ]
                )
            )

            # create trie for partial word matches
            char_trie = CharTrie.fromkeys(hotword_unigrams)
        else:
            # make an unmatchable pattern
            match_ptn = re.compile(r"^\b$")
            # empty trie
            char_trie = CharTrie()
        return cls(match_ptn, char_trie, weight)


class AbstractLanguageModel(abc.ABC):
    @property
    @abc.abstractmethod
    def order(self) -> int:
        """Get the order of the n-gram language model."""
        raise NotImplementedError()

    @abc.abstractmethod
    def get_start_state(self) -> List[kenlm.State]:
        """Get initial lm state."""
        raise NotImplementedError()

    @abc.abstractmethod
    def score_partial_token(self, partial_token: str) -> float:
        """Get partial token score."""
        raise NotImplementedError()

    @abc.abstractmethod
    def score(
        self, prev_state: kenlm.State, word: str, is_last_word: bool = False
    ) -> Tuple[float, kenlm.State]:
        """Score word conditional on previous lm state."""
        raise NotImplementedError()


class LanguageModel(AbstractLanguageModel):
    def __init__(
        self,
        kenlm_model: kenlm.Model,
        unigrams: Optional[Collection[str]] = None,
        alpha: float = DEFAULT_ALPHA,
        beta: float = DEFAULT_BETA,
        unk_score_offset: float = DEFAULT_UNK_LOGP_OFFSET,
        score_boundary: bool = DEFAULT_SCORE_LM_BOUNDARY,
    ) -> None:
        """Language model container class to consolidate functionality.

        Args:
            kenlm_model: instance of kenlm n-gram language model `kenlm.Model`
            unigrams: list of known word unigrams
            alpha: weight for language model during shallow fusion
            beta: weight for length score adjustment of during scoring
            unk_score_offset: amount of log score offset for unknown tokens
            score_boundary: whether to have kenlm respect boundaries when scoring
        """
        self._kenlm_model = kenlm_model
        if unigrams is None:
            logger.warning("No known unigrams provided, decoding results might be a lot worse.")
            unigram_set = set()
            char_trie = None
        else:
            unigram_set = _prepare_unigram_set(unigrams, self._kenlm_model)
            char_trie = CharTrie.fromkeys(unigram_set)
        self._unigram_set = unigram_set
        self._char_trie = char_trie
        self.alpha = alpha
        self.beta = beta
        self.unk_score_offset = unk_score_offset
        self.score_boundary = score_boundary

    @property
    def order(self) -> int:
        """Get the order of the n-gram language model."""
        return cast(int, (self._kenlm_model.order))

    def get_start_state(self) -> kenlm.State:
        """Get initial lm state."""
        start_state = _get_empty_lm_state()
        if self.score_boundary:
            self._kenlm_model.BeginSentenceWrite(start_state)
        else:
            self._kenlm_model.NullContextWrite(start_state)
        return start_state

    def _get_raw_end_score(self, start_state: kenlm.State) -> float:
        """Calculate final lm score."""
        if self.score_boundary:
            end_state = _get_empty_lm_state()
            score: float = self._kenlm_model.BaseScore(start_state, "</s>", end_state)
        else:
            score = 0.0
        return score

    def score_partial_token(self, partial_token: str) -> float:
        """Get partial token score."""
        if self._char_trie is None:
            is_oov = 1.0
        else:
            is_oov = int(self._char_trie.has_node(partial_token) == 0)
        unk_score = self.unk_score_offset * is_oov
        # if unk token length exceeds expected length then additionally decrease score
        if len(partial_token) > AVG_TOKEN_LEN:
            unk_score = unk_score * len(partial_token) / AVG_TOKEN_LEN
        return unk_score

    def score(
        self, prev_state: kenlm.State, word: str, is_last_word: bool = False
    ) -> Tuple[float, kenlm.State]:
        """Score word conditional on start state."""
        end_state = _get_empty_lm_state()
        lm_score = self._kenlm_model.BaseScore(prev_state, word, end_state)
        # override UNK prob. use unigram set if we have because it's faster
        if (
            len(self._unigram_set) > 0
            and word not in self._unigram_set
            or word not in self._kenlm_model
        ):
            lm_score += self.unk_score_offset
        # add end of sentence context if needed
        if is_last_word:
            # note that we want to return the unmodified end_state to keep extension capabilities
            lm_score = lm_score + self._get_raw_end_score(end_state)
        lm_score = self.alpha * lm_score * LOG_BASE_CHANGE_FACTOR + self.beta
        return lm_score, end_state


class MultiLanguageModel(AbstractLanguageModel):
    def __init__(self, language_models: List[LanguageModel]) -> None:
        """Container for multiple language models.

        Args:
            language_models: list of language models
        """
        if len(language_models) < 2:
            raise ValueError("This class is meant to contain at least 2 language models.")
        self._language_models = language_models

    @property
    def order(self) -> int:
        """Get the maximum order of the contained n-gram language model."""
        return max([lm.order for lm in self._language_models])

    def get_start_state(self) -> List[kenlm.State]:
        """Get initial lm state."""
        return [lm.get_start_state() for lm in self._language_models]

    def score_partial_token(self, partial_token: str) -> float:
        """Get partial token score."""
        return float(
            np.mean([lm.score_partial_token(partial_token) for lm in self._language_models])
        )

    def score(
        self, prev_state: List[kenlm.State], word: str, is_last_word: bool = False
    ) -> Tuple[float, List[kenlm.State]]:
        """Score word conditional on previous lm state."""
        score = 0.0
        end_state = []
        for lm_prev_state, lm in zip(prev_state, self._language_models):
            lm_score, lm_end_state = lm.score(lm_prev_state, word, is_last_word=is_last_word)
            score += lm_score
            end_state.append(lm_end_state)
        score = score / len(self._language_models)
        return score, end_state



import numpy as np
import librosa
import math
import platform
import iman.Features.mfcc.getmfcc_from_librosa as libr 

OPERATING_SYSTEM = platform.system()
SYSTEM_ARCH = platform.platform()
SYSTEM_PROC = platform.processor()


if OPERATING_SYSTEM == 'Windows':
    from pyrubberband import pyrb
else:
    from . import pyrb

if OPERATING_SYSTEM == 'Darwin':
    wav_resolution = "polyphase" if SYSTEM_PROC == ARM or ARM in SYSTEM_ARCH else "sinc_fastest" 
    wav_resolution_float_resampling = "kaiser_best" if SYSTEM_PROC == ARM or ARM in SYSTEM_ARCH else wav_resolution 
    is_macos = True
else:
    wav_resolution = "sinc_fastest"
    wav_resolution_float_resampling = wav_resolution 
 
def spectrogram_to_wave_old(spec, hop_length=1024):
    if spec.ndim == 2:
        wave = libr.istft(spec, hop_length=hop_length)
    elif spec.ndim == 3:
        spec_left = np.asfortranarray(spec[0])
        spec_right = np.asfortranarray(spec[1])

        wave_left = libr.istft(spec_left, hop_length=hop_length)
        wave_right = libr.istft(spec_right, hop_length=hop_length)
        wave = np.asfortranarray([wave_left, wave_right])

    return wave

def wave_to_spectrogram_old(wave, hop_length, n_fft):


    if (wave.ndim != 2):
       wave_left = np.asfortranarray(wave)
    
    if (wave.ndim == 2):
        wave_left = np.asfortranarray(wave[0])
        wave_right = np.asfortranarray(wave[1])

    spec_left = libr.STFT(wave_left, n_fft, hop_length=hop_length)
    
    if (wave.ndim == 2):
        spec_right = libr.STFT(wave_right, n_fft, hop_length=hop_length)
    
    if (wave.ndim == 1):
       spec = np.asfortranarray([spec_left,spec_left]) 
    else:
       spec = np.asfortranarray([spec_left, spec_right])    
    

    return spec
 
def convert_channels(spec, mp, band):
    cc = mp['band'][band].get('convert_channels')

    if 'mid_side_c' == cc:
        spec_left = np.add(spec[0], spec[1] * .25)
        spec_right = np.subtract(spec[1], spec[0] * .25)
    elif 'mid_side' == cc:
        spec_left = np.add(spec[0], spec[1]) / 2
        spec_right = np.subtract(spec[0], spec[1])
    elif 'stereo_n' == cc:
        spec_left = np.add(spec[0], spec[1] * .25) / 0.9375
        spec_right = np.add(spec[1], spec[0] * .25) / 0.9375
    else:
        return spec
        
    return np.asfortranarray([spec_left, spec_right])
  
def wave_to_spectrogram(wave, hop_length, n_fft, mp, band, is_v51_model=False):

    if wave.ndim == 1:
        wave = np.asfortranarray([wave,wave])

    if not is_v51_model:
        if mp['reverse']:
            wave_left = np.flip(np.asfortranarray(wave[0]))
            wave_right = np.flip(np.asfortranarray(wave[1]))
        elif mp['mid_side']:
            wave_left = np.asfortranarray(np.add(wave[0], wave[1]) / 2)
            wave_right = np.asfortranarray(np.subtract(wave[0], wave[1]))
        elif mp['mid_side_b2']:
            wave_left = np.asfortranarray(np.add(wave[1], wave[0] * .5))
            wave_right = np.asfortranarray(np.subtract(wave[0], wave[1] * .5))
        else:
            wave_left = np.asfortranarray(wave[0])
            wave_right = np.asfortranarray(wave[1])
    else:
        wave_left = np.asfortranarray(wave[0])
        wave_right = np.asfortranarray(wave[1])

    spec_left = libr.STFT(wave_left, n_fft, hop_length=hop_length)
    spec_right = libr.STFT(wave_right, n_fft, hop_length=hop_length)
    
    spec = np.asfortranarray([spec_left, spec_right])

    if is_v51_model:
        spec = convert_channels(spec, mp, band)

    return spec
	
   
def combine_spectrograms(specs, mp, is_v51_model=False):
    l = min([specs[i].shape[2] for i in specs])    
    spec_c = np.zeros(shape=(2, mp['bins'] + 1, l), dtype=np.complex64)
    offset = 0
    bands_n = len(mp['band'])
    
    for d in range(1, bands_n + 1):
        h = mp['band'][d]['crop_stop'] - mp['band'][d]['crop_start']
        spec_c[:, offset:offset+h, :l] = specs[d][:, mp['band'][d]['crop_start']:mp['band'][d]['crop_stop'], :l]
        offset += h
        
    if offset > mp['bins']:
        raise ValueError('Too much bins')
        
    # lowpass fiter
    
    if mp['pre_filter_start'] > 0:
        if is_v51_model:
            spec_c *= get_lp_filter_mask(spec_c.shape[1], mp['pre_filter_start'], mp['pre_filter_stop'])
        else:
            if bands_n == 1:
                spec_c = fft_lp_filter(spec_c, mp['pre_filter_start'], mp['pre_filter_stop'])
            else:
                gp = 1        
                for b in range(mp['pre_filter_start'] + 1, mp['pre_filter_stop']):
                    g = math.pow(10, -(b - mp['pre_filter_start']) * (3.5 - gp) / 20.0)
                    gp = g
                    spec_c[:, b, :] *= g
                
    return np.asfortranarray(spec_c)
	
def make_padding(width, cropsize, offset):
    left = offset
    roi_size = cropsize - offset * 2
    if roi_size == 0:
        roi_size = cropsize
    right = roi_size - (width % roi_size) + left

    return left, right, roi_size
	

def spectrogram_to_wave(spec, hop_length=1024, mp={}, band=0, is_v51_model=True):
    spec_left = np.asfortranarray(spec[0])
    spec_right = np.asfortranarray(spec[1])
    
    
    
    wave_left = libr.istft(spec_left, hop_length=hop_length)
    wave_right = libr.istft(spec_right, hop_length=hop_length)
    
    if is_v51_model:
        cc = mp['band'][band].get('convert_channels')
        if 'mid_side_c' == cc:
            return np.asfortranarray([np.subtract(wave_left / 1.0625, wave_right / 4.25), np.add(wave_right / 1.0625, wave_left / 4.25)])    
        elif 'mid_side' == cc:
            return np.asfortranarray([np.add(wave_left, wave_right / 2), np.subtract(wave_left, wave_right / 2)])
        elif 'stereo_n' == cc:
            return np.asfortranarray([np.subtract(wave_left, wave_right * .25), np.subtract(wave_right, wave_left * .25)])
    else:
        if mp['reverse']:
            return np.asfortranarray([np.flip(wave_left), np.flip(wave_right)])
        elif mp['mid_side']:
            return np.asfortranarray([np.add(wave_left, wave_right / 2), np.subtract(wave_left, wave_right / 2)])
        elif mp['mid_side_b2']:
            return np.asfortranarray([np.add(wave_right / 1.25, .4 * wave_left), np.subtract(wave_left / 1.25, .4 * wave_right)])
    
    return np.asfortranarray([wave_left, wave_right])
    

def crop_center(h1, h2):
    h1_shape = h1.size()
    h2_shape = h2.size()

    if h1_shape[3] == h2_shape[3]:
        return h1
    elif h1_shape[3] < h2_shape[3]:
        raise ValueError('h1_shape[3] must be greater than h2_shape[3]')

    s_time = (h1_shape[3] - h2_shape[3]) // 2
    e_time = s_time + h2_shape[3]
    h1 = h1[:, :, :, s_time:e_time]

    return h1
    
def fft_hp_filter(spec, bin_start, bin_stop):
    g = 1.0
    for b in range(bin_start, bin_stop, -1):
        g -= 1 / (bin_start - bin_stop)
        spec[:, b, :] = g * spec[:, b, :]
    
    spec[:, 0:bin_stop+1, :] *= 0

    return spec
    
def cmb_spectrogram_to_wave(spec_m, mp, extra_bins_h=None, extra_bins=None, is_v51_model=False):
    bands_n = len(mp['band'])    
    offset = 0

    for d in range(1, bands_n + 1):
        bp = mp['band'][d]
        spec_s = np.ndarray(shape=(2, bp['n_fft'] // 2 + 1, spec_m.shape[2]), dtype=complex)
        h = bp['crop_stop'] - bp['crop_start']
        spec_s[:, bp['crop_start']:bp['crop_stop'], :] = spec_m[:, offset:offset+h, :]
                
        offset += h
        if d == bands_n: # higher
            if extra_bins_h: # if --high_end_process bypass
                max_bin = bp['n_fft'] // 2
                spec_s[:, max_bin-extra_bins_h:max_bin, :] = extra_bins[:, :extra_bins_h, :]
            if bp['hpf_start'] > 0:
                if is_v51_model:
                    spec_s *= get_hp_filter_mask(spec_s.shape[1], bp['hpf_start'], bp['hpf_stop'] - 1)
                else:
                    spec_s = fft_hp_filter(spec_s, bp['hpf_start'], bp['hpf_stop'] - 1)
            if bands_n == 1:
                wave = spectrogram_to_wave(spec_s, bp['hl'], mp, d, is_v51_model)
            else:
                wave = np.add(wave, spectrogram_to_wave(spec_s, bp['hl'], mp, d, is_v51_model))
        else:
            sr = mp['band'][d+1]['sr']
            if d == 1: # lower
                if is_v51_model:
                    spec_s *= get_lp_filter_mask(spec_s.shape[1], bp['lpf_start'], bp['lpf_stop'])
                else:
                    spec_s = fft_lp_filter(spec_s, bp['lpf_start'], bp['lpf_stop'])
                wave = librosa.resample(spectrogram_to_wave(spec_s, bp['hl'], mp, d, is_v51_model), bp['sr'], sr, res_type=wav_resolution)
            else: # mid
                if is_v51_model:
                    spec_s *= get_hp_filter_mask(spec_s.shape[1], bp['hpf_start'], bp['hpf_stop'] - 1)
                    spec_s *= get_lp_filter_mask(spec_s.shape[1], bp['lpf_start'], bp['lpf_stop'])
                else:
                    spec_s = fft_hp_filter(spec_s, bp['hpf_start'], bp['hpf_stop'] - 1)
                    spec_s = fft_lp_filter(spec_s, bp['lpf_start'], bp['lpf_stop'])
                    
                wave2 = np.add(wave, spectrogram_to_wave(spec_s, bp['hl'], mp, d, is_v51_model))
                wave = librosa.resample(wave2, bp['sr'], sr, res_type=wav_resolution)
        
    return wave


def get_lp_filter_mask(n_bins, bin_start, bin_stop):
    mask = np.concatenate([
        np.ones((bin_start - 1, 1)),
        np.linspace(1, 0, bin_stop - bin_start + 1)[:, None],
        np.zeros((n_bins - bin_stop, 1))
    ], axis=0)

    return mask
    
def get_hp_filter_mask(n_bins, bin_start, bin_stop):
    mask = np.concatenate([
        np.zeros((bin_stop + 1, 1)),
        np.linspace(0, 1, 1 + bin_start - bin_stop)[:, None],
        np.ones((n_bins - bin_start - 2, 1))
    ], axis=0)

    return mask


from .lib_v5.vr_network import nets_new
from .lib_v5.vr_network.model_param_init import ModelParameters
import torch
from .lib_v5 import spec_utils
import librosa
from iman import Audio
from iman import gf,PN,PJ,os,np,PM
from tqdm import tqdm

cropsize=256

import warnings
warnings.filterwarnings('ignore')

def loading_mix(X, mp):

    X_wave, X_spec_s = {}, {}
    
    bands_n = len(mp['band'])
    
    for d in range(bands_n, 0, -1):        
        bp = mp['band'][d]

        wav_resolution = 'polyphase'#bp['res_type']
    
        if d == bands_n: # high-end band
            X_wave[d] = X

        else: # lower bands
            X_wave[d] = librosa.resample(X_wave[d+1], mp['band'][d+1]['sr'], bp['sr'], res_type=wav_resolution)
            
        X_spec_s[d] = spec_utils.wave_to_spectrogram(X_wave[d], bp['hl'], bp['n_fft'], mp, band=d, is_v51_model=True)
        
        # if d == bands_n and is_high_end_process:
        #     input_high_end_h = (bp['n_fft']//2 - bp['crop_stop']) + (mp['pre_filter_stop'] - mp['pre_filter_start'])
        #     input_high_end = X_spec_s[d][:, bp['n_fft']//2-input_high_end_h:bp['n_fft']//2, :]

    X_spec = spec_utils.combine_spectrograms(X_spec_s, mp)
    
    del X_wave, X_spec_s

    return X_spec


def Dereverb(pattern , out_fol , sr = 16000, batchsize=16 , device="cuda"  ,model_path=r"C:\UVR-DeEcho-DeReverb.pth"):
   print('Load Model...') 
   mp={'bins': 672, 'unstable_bins': 8, 'reduction_bins': 530, 'band': {1: {'sr': 7350, 'hl': 80, 'n_fft': 640, 'crop_start': 0, 'crop_stop': 85, 'lpf_start': 25, 'lpf_stop': 53, 'res_type': 'polyphase'}, 2: {'sr': 7350, 'hl': 80, 'n_fft': 320, 'crop_start': 4, 'crop_stop': 87, 'hpf_start': 25, 'hpf_stop': 12, 'lpf_start': 31, 'lpf_stop': 62, 'res_type': 'polyphase'}, 3: {'sr': 14700, 'hl': 160, 'n_fft': 512, 'crop_start': 17, 'crop_stop': 216, 'hpf_start': 48, 'hpf_stop': 24, 'lpf_start': 139, 'lpf_stop': 210, 'res_type': 'polyphase'}, 4: {'sr': 44100, 'hl': 480, 'n_fft': 960, 'crop_start': 78, 'crop_stop': 383, 'hpf_start': 130, 'hpf_stop': 86, 'res_type': 'kaiser_fast'}}, 'sr': 44100, 'pre_filter_start': 668, 'pre_filter_stop': 672, 'mid_side': False, 'mid_side_b': False, 'mid_side_b2': False, 'stereo_w': False, 'stereo_n': False, 'reverse': False}
   PM(out_fol)
   cpu = torch.device('cpu')
   nout, nout_lstm = 64, 128
   n_fft = mp['bins'] * 2
   model = nets_new.CascadedNet(n_fft, nout=nout, nout_lstm=nout_lstm)
   try:
      model.load_state_dict(torch.load(model_path, map_location=cpu))
   except:
      print("WARNING: Download model from 'https://huggingface.co/seanghay/uvr_models/resolve/main/UVR-DeEcho-DeReverb.pth' and then set model_path")   
      return
   model.to(device)
   
   files = gf(pattern)
   print('Find ' +  str(len(files)) + ' files.')
   print('Batch size= ' + str(batchsize))
   print('Sampling Rate= ' + str(sr))
   print('Start Dereverbing...')
   print('-----------------------------------------------------------------')
   for fname in files:
    print(fname)
    # XXX = librosa.load(fname, mono=False, sr=sr)[0]
    XXX = Audio.Read(fname , sr=sr,mono=False)
    chh=1 
    _w1=[]
    _w2=[]    
    for X in (XXX):
       print('Channel ' + str(chh))   
       chh=chh+1
       X_spec = loading_mix(X.T, mp)
       X_mag = np.abs(X_spec)
       X_phase = np.angle(X_spec)

       n_frame = X_mag.shape[2]
       pad_l, pad_r, roi_size = spec_utils.make_padding(n_frame, cropsize, model.offset)
       
       X_mag_pad = np.pad(X_mag, ((0, 0), (0, 0), (pad_l, pad_r)), mode='constant')
       X_mag_pad /= X_mag_pad.max()
       
       X_dataset = []
       patches = (X_mag_pad.shape[2] - 2 * model.offset) // roi_size
       
       for i in range(patches):
               start = i * roi_size
               X_mag_crop = X_mag_pad[:, :, start:start + cropsize]
               X_dataset.append(X_mag_crop)
               
       X_dataset = np.asarray(X_dataset)
       
       model.eval()        
       with torch.no_grad():
               mask = []
               # To reduce the overhead, dataloader is not used.
               for i in tqdm(range(0, patches, batchsize)):
                   X_batch = X_dataset[i: i + batchsize]
                   X_batch = torch.from_numpy(X_batch).to(device)
       
                   pred = model.predict_mask(X_batch)
       
                   pred = pred.detach().cpu().numpy()
                   pred = np.concatenate(pred, axis=2)
                   mask.append(pred)
       
               mask = np.concatenate(mask, axis=2)
       mask = mask[:, :, :n_frame]
       
       v_spec = mask * X_mag * np.exp(1.j * X_phase)
       y_spec = (1 - mask) * X_mag * np.exp(1.j * X_phase)
       
       wave = spec_utils.cmb_spectrogram_to_wave(v_spec, mp, is_v51_model=True).T
       wave_2 = spec_utils.cmb_spectrogram_to_wave(y_spec, mp, is_v51_model=True).T

       _w1.append(wave[:,0])
       _w2.append(wave_2[:,0])
       
       

    
    Audio.Write(PJ(out_fol , PN(fname) + "_1.wav") ,np.array(_w1).T ,sr )
    print('out_path= ' + PJ(out_fol , PN(fname) + "_1.wav"))
    Audio.Write(PJ(out_fol , PN(fname) + "_2.wav") ,np.array(_w2).T ,sr )
    print('out_path= ' + PJ(out_fol , PN(fname) + "_2.wav"))
        
        

def Denoise(pattern , out_fol , sr = 16000, batchsize=16 , device="cuda"  ,model_path=r"C:\UVR-DeNoise-Lite.pth"):
   print('Load Model...') 
   PM(out_fol)
   cpu = torch.device('cpu')
   nout, nout_lstm = 16, 128
   hop_length=1024
   n_fft = 2048
   
   model = nets_new.CascadedNet(n_fft, nout=nout, nout_lstm=nout_lstm)
   try:
      model.load_state_dict(torch.load(model_path, map_location=cpu))
   except:
      print("WARNING: Download model from 'https://huggingface.co/seanghay/uvr_models/resolve/main/UVR-DeNoise-Lite.pth' and then set model_path")   
      return
   model.to(device)
   
   files = gf(pattern)
   print('Find ' +  str(len(files)) + ' files.')
   print('Batch size= ' + str(batchsize))
   print('Sampling Rate= ' + str(sr))
   print('Start DeNoising...')
   print('-----------------------------------------------------------------')
   for fname in files:
       print(fname)
       # X = Audio.Read(fname , sr , mono=False)
       X = librosa.load(fname, mono=False, sr=sr)[0]
       
       X_spec = spec_utils.wave_to_spectrogram_old(X, hop_length, n_fft)
       
      
       X_mag = np.abs(X_spec)
       X_phase = np.angle(X_spec)


       n_frame = X_mag.shape[2]
       pad_l, pad_r, roi_size = spec_utils.make_padding(n_frame, cropsize, model.offset)
       
       X_mag_pad = np.pad(X_mag, ((0, 0), (0, 0), (pad_l, pad_r)), mode='constant')
       X_mag_pad /= X_mag_pad.max()
       
       X_dataset = []
       patches = (X_mag_pad.shape[2] - 2 * model.offset) // roi_size
       
       for i in range(patches):
               start = i * roi_size
               X_mag_crop = X_mag_pad[:, :, start:start + cropsize]
               X_dataset.append(X_mag_crop)
               
       X_dataset = np.asarray(X_dataset)
       
       model.eval()        
       with torch.no_grad():
               mask = []
               # To reduce the overhead, dataloader is not used.
               for i in tqdm(range(0, patches, batchsize)):
                   X_batch = X_dataset[i: i + batchsize]
                   X_batch = torch.from_numpy(X_batch).to(device)
       
                   pred = model.predict_mask(X_batch)
       
                   pred = pred.detach().cpu().numpy()
                   pred = np.concatenate(pred, axis=2)
                   mask.append(pred)
       
               mask = np.concatenate(mask, axis=2)
       mask = mask[:, :, :n_frame]
       
       
       v_spec = (1 - mask) * X_mag * np.exp(1.j * X_phase)
       wave = spec_utils.spectrogram_to_wave_old(v_spec, hop_length=1024)
       
       import scipy.io.wavfile as __wave    
       
       if (X.ndim==2):
           xx = np.asarray( [np.int16(wave[0]*32768) , np.int16(wave[1]*32768)]).transpose()
       else:
           xx = np.int16(wave[0]*32768)       

       __wave.write(PJ(out_fol , PN(fname) + ".wav") , sr, xx)

       print('out_path= ' + PJ(out_fol , PN(fname) + ".wav"))
    


import os
import numpy as np
from iman import Audio
from iman import Features


def media2feats(wavname,sr=8000,start_from=0 ,dur=-1,ffmpeg_path='c:\\ffmpeg.exe'):
    try:
       sig1 = Audio.Read(wavname ,sr,start_from , dur ,True, ffmpeg_path)
    except :
        return [],[]
    sig=sig1+1e-7
    try:    
      mfccs,loge  = Features.LS.Get(sig , sr , True)   
    except:
      print('error in mfcc extraction')
      return [] ,[]   
    if (len(loge)<68):   
           return [] ,[]
    return  mfccs , loge     
  
